from __future__ import annotations

import argparse
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

import yaml
from yamlinclude import YamlIncludeConstructor

from puma.hlplots import Results, Tagger, combine_suffixes, get_included_taggers
from puma.hlplots.yutils import get_tagger_name
from puma.utils import logger

ALL_PLOTS = ["roc", "scan", "disc", "probs", "peff"]


def get_args(args):
    parser = argparse.ArgumentParser(description="YUMA: Plotting from Yaml in pUMA")
    parser.add_argument("-c", "--config", required=True, type=str, help="Path to config")

    parser.add_argument(
        "-p",
        "--plots",
        nargs="+",
        choices=ALL_PLOTS,
        help=f"Plot types to make. Allowed are: {ALL_PLOTS} ",
    )
    parser.add_argument(
        "-s",
        "--signals",
        nargs="+",
        choices=["bjets", "cjets"],
        help="Signals to plot",
    )
    parser.add_argument("-d", "--dir", type=Path, help="Base sample directory")

    return parser.parse_args(args)


@dataclass
class YumaConfig:
    config_path: Path
    plot_dir: Path

    results_config: dict[str, dict[str, str]]
    taggers_config: dict
    taggers: list[str] | list[Tagger] | None = None

    timestamp: bool = False
    base_path: Path = None

    # dict like {roc : [list of roc plots], scan: [list of scan plots], ...}
    plots: dict[list[dict[str, dict[str, str]]]] = field(default_factory=list)

    def __post_init__(self):
        # Define a plot directory based on the plot config file name, and a date time
        plot_dir_name = self.config_path.stem
        if self.timestamp:
            date_time_file = datetime.now().strftime("%Y%m%d_%H%M%S")
            plot_dir_name += "_" + date_time_file
        self.plot_dir_final = Path(self.plot_dir) / plot_dir_name

        for k, kwargs in self.taggers_config.items():
            kwargs["yaml_name"] = k
        if not self.taggers:
            logger.info("No taggers specified in config, using all")
            self.taggers = list(self.taggers_config.keys())

    @classmethod
    def load_config(cls, path: Path, **kwargs) -> YumaConfig:
        if not path.exists():
            raise FileNotFoundError(f"Config at {path} does not exist")
        with open(path) as f:
            config = yaml.safe_load(f)

        config = cls(config_path=path, **config, **kwargs)
        config.check_config()
        return config

    def check_config(self):
        """Checks the config for any issues, raises an error if any are found."""
        allowed_keys = ["signal", "plot_kwargs", "include_taggers", "exclude_taggers", "reference"]
        for plots in self.plots.values():
            for p in plots:
                for k in p:
                    if k not in allowed_keys:
                        raise ValueError(
                            f"Unknown key {k} in plot config. Maybe '{k}' belongs"
                            "under 'plot_kwargs'?"
                        )

        return True

    def get_results(self):
        """
        Create the high-level 'Results' object from the config file, using the
        previously set signal and sample. Iterates and loads all models in the config
        file, and adds them.
        """
        kwargs = self.results_config
        kwargs["signal"] = self.signal
        kwargs["perf_vars"] = self.peff_vars

        sample_path = kwargs.pop("sample_path", None)
        if self.base_path and sample_path:
            sample_path = self.base_path / sample_path

        # Instantiate the results object
        results = Results(**kwargs, output_dir=self.plot_dir_final)

        # Add taggers to results, then bulk load
        for key, t in self.taggers_config.items():
            if key not in self.taggers:
                continue
            # if the a sample is not defined for the tagger, use the default sample
            if not sample_path and not t.get("sample_path", None):
                raise ValueError(f"No sample path defined for tagger {key}")
            if sample_path and not t.get("sample_path", None):
                t["sample_path"] = sample_path
            if self.base_path and t.get("sample_path", None):
                t["sample_path"] = self.base_path / t["sample_path"]
            # Allows automatic selection of tagger name in eval files
            t["name"] = get_tagger_name(
                t.get("name", None), t["sample_path"], key, results.flavours
            )

            results.add(Tagger(**t))

        results.load()
        self.results = results

    @property
    def signals(self):
        """Iterates all plots in the config and returns a list of all signals."""
        return sorted({p["signal"] for pt in self.plots.values() for p in pt})

    @property
    def peff_vars(self):
        """Iterates plots and returns a list of all performance variables."""
        return list({p["plot_kwargs"].get("perf_var", "pt") for p in self.plots.get("peff", [])})

    def make_plots(self, plot_types):
        """Makes all desired plots.

        Parameters
        ----------
        plot_types : list[str]
            List of plot types to make.
        """
        for plot_type, plots in self.plots.items():
            if plot_type not in plot_types:
                continue
            for plot in plots:
                if plot["signal"] != self.signal:
                    continue
                self.results.taggers, all_taggers, inc_str = get_included_taggers(
                    self.results, plot
                )
                plot_kwargs = plot.get("plot_kwargs", {})
                plot_kwargs["suffix"] = combine_suffixes([plot_kwargs.get("suffix", ""), inc_str])
                self.results.make_plot(plot_type, plot_kwargs)
                self.results.taggers = all_taggers


def main(args=None):
    args = get_args(args)

    config_path = Path(args.config)
    YamlIncludeConstructor.add_to_loader_class(
        loader_class=yaml.SafeLoader, base_dir=config_path.parent
    )
    yuma = YumaConfig.load_config(config_path, base_path=args.dir)

    # select and check plots
    plots = args.plots if args.plots else ALL_PLOTS
    if missing := [p for p in plots if p not in ALL_PLOTS]:
        raise ValueError(f"Unknown plot types {missing}, choose from {ALL_PLOTS}")

    # select and check signals
    signals = args.signals if args.signals else yuma.signals
    if missing := [s for s in signals if s not in yuma.signals]:
        raise ValueError(f"Unknown signals {missing}, choose from {yuma.signals}")

    logger.info(f"Plotting in {yuma.plot_dir_final}")

    logger.info("Instantiating Results")
    yuma.signal = signals[0]
    yuma.get_results()  # only run once
    for signal in signals:
        logger.info(f"Plotting signal {signal}")
        yuma.signal = signal
        yuma.results.set_signal(signal)
        yuma.make_plots(plots)


if __name__ == "__main__":
    main()



"""Configuration for logger of puma."""

from __future__ import annotations

import logging
import os
from typing import ClassVar


class CustomFormatter(logging.Formatter):
    """
    Logging Formatter to add colours and count warning / errors using implementation
    from
    https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output.
    """

    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    green = "\x1b[32;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    debugformat = "%(asctime)s - %(levelname)s:%(name)s: %(message)s (%(filename)s:%(lineno)d)"
    date_format = "%(levelname)s:%(name)s: %(message)s"

    formats: ClassVar = {
        logging.DEBUG: grey + debugformat + reset,
        logging.INFO: green + date_format + reset,
        logging.WARNING: yellow + date_format + reset,
        logging.ERROR: red + debugformat + reset,
        logging.CRITICAL: bold_red + debugformat + reset,
    }

    def format(self, record):
        log_fmt = self.formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_log_level(
    level: str,
):
    """Get logging levels with string key.

    Parameters
    ----------
    level : str
        Log level as string.

    Returns
    -------
    logging level
        logging object with log level info

    Raises
    ------
    ValueError
        If non-valid option is given
    """
    log_levels = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }
    if level not in log_levels:
        raise ValueError(f"The 'DebugLevel' option {level} is not valid.")
    return log_levels[level]


def initialise_logger(
    log_level: str | None = None,
):
    """Initialise.

    Parameters
    ----------
    log_level : str, optional
        Logging level defining the verbose level. Accepted values are:
        CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET, by default None
        If the log_level is not set, it will

    Returns
    -------
    logger
        logger object with new level set
    """
    log_level = get_log_level(
        os.environ.get("LOG_LEVEL", "INFO") if log_level is None else log_level
    )

    puma_logger = logging.getLogger("puma")
    puma_logger.setLevel(log_level)
    ch_handler = logging.StreamHandler()
    ch_handler.setLevel(log_level)
    ch_handler.setFormatter(CustomFormatter())

    puma_logger.addHandler(ch_handler)
    puma_logger.propagate = False
    return puma_logger


def set_log_level(
    puma_logger,
    log_level: str,
):
    """Setting log level.

    Parameters
    ----------
    puma_logger : logger
        logger object
    log_level : str
        Logging level corresponding CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET
    """
    puma_logger.setLevel(get_log_level(log_level))
    for handler in puma_logger.handlers:
        handler.setLevel(get_log_level(log_level))


logger = initialise_logger()



from __future__ import annotations

import numpy as np
import scipy as scp


def confusion_matrix(
    targets: np.ndarray,
    predictions: np.ndarray,
    sample_weights: np.ndarray | None = None,
    normalize: str | None = "rownorm",
) -> np.ndarray:
    """
    Parameters
    ----------
    targets : 1d np.ndarray
        target labels
    predictions : 1d np.ndarray
        predicted labels (output of the classifier)
    sample_weights : np.ndarray, optional
        Weight of each sample; if None, each sample weights the same. Defaults to None.
    normalize : str | None, optional
        Normalization of the confusion matrix. Can be:
        None : Give raw counts;
        "rownorm": Normalize across the prediction class, i.e. such that the rows add to one;
        "colnorm": Normalize across the target class, i.e. such that the columns add to one;
        "all" : Normalize across all examples, i.e. such that all matrix entries add to one.
        Defaults to "rownorm".

    Returns
    -------
    np.ndarray : the confusion matrix.

    Example
    --------
    >>> targets = np.array([2, 0, 2, 2, 0, 1])
    >>> predictions = np.array([0, 0, 2, 2, 0, 2])
    >>> weights = np.array([1, 0.5, 0.5, 1, 0.2, 1])
    >>> confusion_matrix(targets, predictions, sample_weights=weights)
    np.array([[1.  0.  0. ]
        [0.  0.  1. ]
        [0.4 0.  0.6]])
    """
    # Checking that targets and predictions have the same sample size
    assert (
        targets.shape[0] == predictions.shape[0]
    ), "confusion_matrix: Predictions and targets must have the same sample size"
    # If user gives samples' weights, check that the sample size is consistent with the labels
    if sample_weights is not None:
        assert (
            sample_weights.shape[0] == targets.shape[0]
        ), "confusion_matrix: Mismatch between targets' and sample weights' size"

    if normalize is not None:
        assert normalize in {
            "rownorm",
            "colnorm",
            "all",
        }, "confusion_matrix: invalid normalization keyword"

    # Finding number of target classes
    # (i.e. max value of the categorical indexing plus one,
    # since categorical index starts from zero)
    n_classes = int(np.max(targets)) + 1

    # If no samples' weights are given, give to each sample weight = 1
    if sample_weights is None:
        sample_weights = np.ones_like(targets)

    # Calculate the raw count Confusion Matrix
    cm = scp.sparse.coo_matrix(
        (sample_weights, (targets, predictions)),
        shape=(n_classes, n_classes),
        dtype="float",
    ).toarray()

    # Eventually normalize the Confusion Matrix
    with np.errstate(all="warn"):
        if normalize == "all":
            cm /= cm.sum()
        elif normalize == "rownorm":
            cm /= cm.sum(axis=1, keepdims=True)
        elif normalize == "colnorm":
            cm /= cm.sum(axis=0, keepdims=True)

    # Returning the CM with nan converted to zero
    return np.nan_to_num(cm)



from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.testing.compare import compare_images

from puma.matshow import MatshowPlot
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class TestMatshowPlot(unittest.TestCase):
    def setUp(self):
        # Set up directories for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

    def test_matrix(self):
        # test matrix
        mat = np.array([
            [0.55136792, 0.58809975, 0.43140183],
            [0.83374029, 0.24685411, 0.87419068],
            [0.08393056, 0.3477517, 0.81477693],
            [0.39839215, 0.54854937, 0.48571167],
        ])
        plot_mat = MatshowPlot(colormap=plt.cm.PiYG, x_ticks_rotation=0)
        name = "test_matrix.png"
        plot_mat.draw(mat)
        plot_mat.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_matrix_nocbar(self):
        # test matrix
        mat = np.array([
            [0.55136792, 0.58809975, 0.43140183],
            [0.83374029, 0.24685411, 0.87419068],
            [0.08393056, 0.3477517, 0.81477693],
            [0.39839215, 0.54854937, 0.48571167],
        ])
        plot_mat = MatshowPlot(colormap=plt.cm.PiYG, x_ticks_rotation=0, show_cbar=False)
        name = "test_matrix_nocbar.png"
        plot_mat.draw(mat)
        plot_mat.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_matrix_fully_customized(self):
        # test matrix
        mat = np.array([
            [0.55136792, 0.58809975, 0.43140183],
            [0.83374029, 0.24685411, 0.87419068],
            [0.08393056, 0.3477517, 0.81477693],
            [0.39839215, 0.54854937, 0.48571167],
        ])
        x_ticks = ["a", "b", "c"]
        y_ticks = ["d", "e", "f", "g"]

        plot_mat = MatshowPlot(
            x_ticklabels=x_ticks,
            x_ticks_rotation=45,
            y_ticklabels=y_ticks,
            show_entries=True,
            show_percentage=True,
            text_color_threshold=0.6,
            cbar_label="Scalar values",
        )
        plot_mat.draw(mat)
        name = "test_matrix_fully_customized.png"
        plot_mat.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_matrix_fully_customized_no_entries(self):
        # test matrix
        mat = np.array([
            [0.55136792, 0.58809975, 0.43140183],
            [0.83374029, 0.24685411, 0.87419068],
            [0.08393056, 0.3477517, 0.81477693],
            [0.39839215, 0.54854937, 0.48571167],
        ])
        x_ticks = ["a", "b", "c"]
        y_ticks = ["d", "e", "f", "g"]

        plot_mat = MatshowPlot(
            x_ticklabels=x_ticks,
            x_ticks_rotation=45,
            y_ticklabels=y_ticks,
            show_entries=False,
            show_percentage=True,
            text_color_threshold=0.6,
            cbar_label="Scalar values",
        )
        name = "test_matrix_fully_customized_no_entries.png"
        plot_mat.draw(mat)
        plot_mat.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )



"""Unit test script for the functions in hlplots/n_track_origin.py."""

from __future__ import annotations

import os

# Uncomment line below to update expected image
# import shutil
import tempfile
import unittest
from urllib.request import urlretrieve

import numpy as np
from ftag import Flavours
from matplotlib.testing.compare import compare_images

from puma.hlplots import n_tracks_per_origin
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class NTrackOriginTestCase(unittest.TestCase):
    """Test class for the function."""

    def setUp(self):
        """Create a default dataset for testing."""
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.data_url = "https://umami-ci-provider.web.cern.ch/plot_input_vars/"
        self.r22_url = os.path.join(self.data_url, "plot_input_vars_r22_check.h5")
        self.r22_test_file = os.path.join(self.actual_plots_dir, "plot_input_vars_r22_check.h5")
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")
        urlretrieve(self.r22_url, self.r22_test_file)  # noqa: S310

    def test_one_sample_all_flavour(self):
        """Test the one_sample_all_flavour plot."""
        n_tracks_per_origin(
            flavour_list=[
                Flavours.bjets,
                Flavours.cjets,
                Flavours.ujets,
            ],
            files={
                "ttbar": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(20_000, 250_000, 20),
                    "process_label": "$t\\bar{t}$",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
            },
            plot_path=self.actual_plots_dir,
            plot_type="one_sample_all_flavour",
            track_origin_dict=None,
            plot_format="png",
        )

        # Define a plot name
        name = "ttbar_all_flavour.png"

        # Uncomment line below to update expected image
        # shutil.copy(
        #     f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}"
        # )

        # Check
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_all_samples_one_flavour(self):
        """Test the all_samples_one_flavour plot."""
        n_tracks_per_origin(
            flavour_list=[
                Flavours.bjets,
                Flavours.cjets,
                Flavours.ujets,
            ],
            files={
                "ttbar_1": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(20_000, 100_000, 20),
                    "process_label": "$t\\bar{t}$ 1",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
                "ttbar_2": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(100_000, 250_000, 20),
                    "process_label": "$t\\bar{t}$ 2",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
            },
            plot_path=self.actual_plots_dir,
            plot_type="all_samples_one_flavour",
            track_origin_dict=None,
            plot_format="png",
        )

        # Define a plot name
        names = [
            "bjets_all_samples.png",
            "cjets_all_samples.png",
            "ujets_all_samples.png",
        ]

        # Check all plots
        for name in names:
            # Uncomment line below to update expected image
            # shutil.copy(
            #     f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}"
            # )

            # Check
            self.assertIsNone(
                compare_images(
                    f"{self.actual_plots_dir}/{name}",
                    f"{self.expected_plots_dir}/{name}",
                    tol=1,
                )
            )

    def test_extra_kwargs_all_flavour(self):
        """Test extra kwargs."""
        n_tracks_per_origin(
            flavour_list=[
                Flavours.bjets,
                Flavours.cjets,
                Flavours.ujets,
            ],
            files={
                "ttbar_test_kwargs": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(20_000, 250_000, 20),
                    "process_label": "$t\\bar{t}$",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
            },
            plot_path=self.actual_plots_dir,
            plot_type="one_sample_all_flavour",
            track_origin_dict=None,
            plot_format="png",
            atlas_second_tag="Test extra kwargs",
        )

        # Define a plot name
        name = "ttbar_test_kwargs_all_flavour.png"

        # Uncomment line below to update expected image
        # shutil.copy(
        #     f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}"
        # )

        # Check
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_extra_kwargs_all_samples(self):
        """Test extra kwargs."""
        n_tracks_per_origin(
            flavour_list=[
                Flavours.bjets,
                Flavours.cjets,
                Flavours.ujets,
            ],
            files={
                "ttbar_1": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(20_000, 100_000, 20),
                    "process_label": "$t\\bar{t}$ 1",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
                "ttbar_2": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(100_000, 250_000, 20),
                    "process_label": "$t\\bar{t}$ 2",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
            },
            plot_path=self.actual_plots_dir,
            plot_type="all_samples_one_flavour",
            plot_name="extra_kwargs",
            track_origin_dict=None,
            plot_format="png",
            atlas_second_tag="Test extra kwargs",
        )

        # Define a plot name
        names = [
            "extra_kwargs_bjets_all_samples.png",
            "extra_kwargs_cjets_all_samples.png",
            "extra_kwargs_ujets_all_samples.png",
        ]

        # Check all plots
        for name in names:
            # Uncomment line below to update expected image
            # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

            # Check
            self.assertIsNone(
                compare_images(
                    f"{self.actual_plots_dir}/{name}",
                    f"{self.expected_plots_dir}/{name}",
                    tol=1,
                )
            )

    def test_custom_track_origins(self):
        """Test custom grouping of track origins."""
        n_tracks_per_origin(
            flavour_list=[
                Flavours.bjets,
                Flavours.cjets,
                Flavours.ujets,
            ],
            files={
                "ttbar_test_custom_track_origins": {
                    "filepath": self.r22_test_file,
                    "n_jets": 10000,
                    "tracks_name": "tracks_loose",
                    "pt_bins": np.linspace(20_000, 250_000, 20),
                    "process_label": "$t\\bar{t}$",
                    "jet_pt_variable": "pt_btagJes",
                    "track_truth_variable": "truthOriginLabel",
                    "flavour_label_variable": "HadronConeExclTruthLabelID",
                },
            },
            plot_path=self.actual_plots_dir,
            plot_type="one_sample_all_flavour",
            track_origin_dict={
                "All": range(8),
                "Fragmentation": [1, 2],
                "HF decay": [3, 4, 5],
                "From $\\tau$": [6],
                "Others": [7],
            },
            plot_format="png",
        )

        # Define a plot name
        name = "ttbar_test_custom_track_origins_all_flavour.png"

        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        # Check
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_wrong_plot_type(self):
        """Test wrong plot type error raising."""
        with self.assertRaises(ValueError):
            n_tracks_per_origin(
                flavour_list=[
                    Flavours.bjets,
                    Flavours.cjets,
                    Flavours.ujets,
                ],
                files={
                    "ttbar_test_custom_track_origins": {
                        "filepath": self.r22_test_file,
                        "n_jets": 10000,
                        "tracks_name": "tracks_loose",
                        "pt_bins": np.linspace(20_000, 250_000, 20),
                        "process_label": "$t\\bar{t}$",
                        "jet_pt_variable": "pt_btagJes",
                        "track_truth_variable": "truthOriginLabel",
                        "flavour_label_variable": "HadronConeExclTruthLabelID",
                    },
                },
                plot_path=self.actual_plots_dir,
                plot_type="crash",
                track_origin_dict=None,
                plot_format="png",
            )



from __future__ import annotations

import unittest

import numpy as np

from puma.utils.logging import logger, set_log_level
from puma.utils.truth_hadron import AssociateTracksToHadron, GetOrderedHadrons, SelectHadron

set_log_level(logger, "DEBUG")


class TruthHadronTestCase(unittest.TestCase):
    """Test class for the puma.truth_hadron functions."""

    def setUp(self):
        self.hadron_barcode = np.array([[1373, 1390, 1375, -1, -1]])
        self.parent_barcode = np.array([[-1, 1373, 1373, -1, -1]])
        self.track_parent = np.array([
            [
                1390,
                1375,
                1390,
                1375,
                -2,
                -2,
                -2,
                -2,
                -2,
                -2,
                -2,
                1375,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ]
        ])
        self.hadron_mask = np.array([[1], [1], [1], [0], [0]])

        self.hadrons = np.array([
            [
                (521, 1373, -1, 47603.895),
                (-411, 1390, 1373, 29951.072),
                (431, 1375, 1373, 15795.845),
                (-1, -1, -1, np.nan),
                (-1, -1, -1, np.nan),
            ]
        ])

    def test_hadron_order(self):
        result = GetOrderedHadrons(self.hadron_barcode, self.parent_barcode, n_max_showers=1)
        known_result = [[[0, 1, 2, -1, -1]]]

        assert (result == known_result).all(), "Test failed: result does not match known_result"
        print("Test passed!")

    def test_track_to_hadron(self):
        result_a, result_b, result_c = AssociateTracksToHadron(
            self.track_parent, self.hadron_barcode, self.hadron_mask
        )

        known_result_a = np.array([
            [
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            ],
            [
                [
                    1,
                    0,
                    1,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            ],
            [
                [
                    0,
                    1,
                    0,
                    1,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    1,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            ],
            [
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            ],
            [
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            ],
        ])

        known_result_b = np.array([
            [
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ]
        ])

        assert (result_a == known_result_a).all(), "Test failed: result does not match known_result"
        print("Test passed!")

        assert (result_b == known_result_b).all(), "Test failed: result does not match known_result"
        print("Test passed!")

        assert (result_c == known_result_b).all(), "Test failed: result does not match known_result"
        print("Test passed!")

        def test_select_hadron(self):
            result = SelectHadron(self.hadrons, np.array([[1]]))

            known_result = np.array([[[-411, 1390, 1373, 29951.072]]])

            assert (result == known_result).all(), "Test failed: result does not match known_result"
            print("Test passed!")



"""Tagger module for high level API."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
from ftag import Cuts, Flavours, Label, get_discriminant

from puma.utils.aux import get_aux_labels
from puma.utils.vertexing import clean_reco_vertices, clean_truth_vertices


@dataclass
class Tagger:
    """Class storing information and results for a tagger."""

    # commonly passed to the constructor
    name: str
    label: str = None
    reference: bool = False
    colour: str = None
    fxs: dict[str, float] = field(default_factory=lambda: {"fc": 0.1, "fb": 0.2})
    aux_tasks: list = field(default_factory=lambda: list(get_aux_labels().keys()))
    sample_path: Path = None

    # this is only read by the Results class
    cuts: Cuts | list | None = None

    # commonly set by the Results or AuxResults class
    scores: np.ndarray = None
    labels: np.ndarray = None
    aux_scores: dict = None
    aux_labels: dict = None
    perf_vars: dict = None
    aux_perf_vars: dict = None
    output_flavours: list = field(
        default_factory=lambda: [Flavours.ujets, Flavours.cjets, Flavours.bjets]
    )
    disc_cut: float = None
    working_point: float = None

    # Used only by YUMA
    yaml_name: str = None

    def __post_init__(self):
        if self.label is None:
            self.label = self.name
        if isinstance(self.cuts, list):
            self.cuts = Cuts.from_list(self.cuts)
        if self.aux_tasks is not None:
            self.aux_scores = dict.fromkeys(self.aux_tasks)
            self.aux_labels = dict.fromkeys(self.aux_tasks)
        if self.sample_path is not None:
            self.sample_path = Path(self.sample_path)

    def __repr__(self):
        return f"{self.name} ({self.label})"

    def is_flav(self, flavour: Label | str):
        """Return indices of jets of given flavour.

        Parameters
        ----------
        flavour : Label | str
            Flavour label to select

        Returns
        -------
        np.ndarray
            Array of indices of the given flavour
        """
        flavour = Flavours[flavour]
        return flavour.cuts(self.labels).idx

    @property
    def probabilities(self):
        """Return the probabilities of the tagger.

        Returns
        -------
        list
            List of probability names
        """
        return [flav.px for flav in self.output_flavours]

    @property
    def variables(self):
        """Return a list of the outputs of the tagger.

        Returns
        -------
        list
            List of the outputs variable names of the tagger
        """
        return [f"{self.name}_{prob}" for prob in self.probabilities]

    @property
    def aux_variables(self):
        """Return a dict of the auxiliary outputs of the tagger for each task.

        Returns
        -------
        aux_outputs: dict
            Dictionary of auxiliary output variables of the tagger

        Raises
        ------
        ValueError
            If element in self.aux_tasks is unrecognized
        """
        aux_outputs = {}

        for aux_type in self.aux_tasks:
            if aux_type == "vertexing":
                if self.name in {"SV1", "JF"}:
                    aux_outputs[aux_type] = f"{self.name}VertexIndex"
                else:
                    aux_outputs[aux_type] = f"{self.name}_aux_VertexIndex"
            elif aux_type == "track_origin":
                aux_outputs[aux_type] = f"{self.name}_aux_TrackOrigin"
            else:
                raise ValueError(f"{aux_type} is not a recognized aux task.")

        return aux_outputs

    def extract_tagger_scores(
        self, source: object, source_type: str = "data_frame", key: str | None = None
    ):
        """Extract tagger scores from data frame or file.

        Parameters
        ----------
        source : object
            pd.DataFrame or file path to h5 file containing pd.DataFrame or structured
            numpy array
        source_type : str, optional
            Indicates from which source scores should be extracted. Possible options are
            `data_frame` when passing a pd.DataFrame, `data_frame_path` when passing a
            file path to a h5 file with a pd.DataFrame, `h5_file` when
            passing a file path to a h5 file with a structured numpy array, or
            `strucuted_array` when passing a structured numpy array,
            by default "data_frame"
        key : str, optional
            Key within h5 file, needs to be provided when using the `source_type`
            `data_frame_path` or `numpy_structured`, by default None

        Raises
        ------
        ValueError
            if source_type is wrongly specified
        """
        if source_type == "data_frame":
            self.scores = source[self.variables].to_records(index=False)
            return
        if source_type == "structured_array":
            self.scores = source[self.variables]
            return
        if key is None:
            raise ValueError(
                "When using a `source_type` other than `data_frame`, you need to"
                " specify the `key`."
            )
        if source_type == "data_frame_path":
            df_in = pd.read_hdf(source, key=key)
            self.scores = df_in[self.variables].to_records(index=False)

        elif source_type == "h5_file":
            with h5py.File(source, "r") as f_h5:
                self.scores = f_h5[key].fields(self.variables)[:]

        else:
            raise ValueError(f"{source_type} is not a valid value for `source_type`.")

    def n_jets(self, flavour: Label | str):
        """Retrieve number of jets of a given flavour.

        Parameters
        ----------
        flavour : Label | str
            Flavour of jets to count

        Returns
        -------
        int
            Number of jets of given flavour
        """
        flavour = Flavours[flavour]
        return len(flavour.cuts(self.labels).values)

    def probs(self, prob_flavour: Label, label_flavour: Label = None):
        """Retrieve probabilities for a given flavour.

        Parameters
        ----------
        prob_flavour : Label
            Return probabilities for this flavour class
        label_flavour : Label, optional
            Only return jets of the given truth flavour, by default None

        Returns
        -------
        np.ndarray
            Probabilities for given flavour
        """
        prob_flavour = Flavours[prob_flavour]
        scores = self.scores
        if label_flavour is not None:
            scores = scores[self.is_flav(label_flavour)]
        return scores[f"{self.name}_{prob_flavour.px}"]

    def discriminant(self, signal: Label, fxs: dict | None = None):
        """Retrieve the discriminant for a given signal class.

        Parameters
        ----------
        signal : Label
            Signal class for which the discriminant should be retrieved
        fxs : dict, optional
            dict of fractions to use instead of the default ones, by default None

        Returns
        -------
        np.ndarray
            Discriminant for given signal class
        """
        signal = Flavours[signal]
        if fxs is None:
            fxs = self.fxs
        fxs = {k: v for k, v in fxs.items() if k != signal.frac_str}
        return get_discriminant(self.scores, self.name, signal, **fxs)

    def vertex_indices(self, incl_vertexing=False):
        """Retrieve cleaned vertex indices for the tagger.

        Parameters
        ----------
        incl_vertexing : bool, optional
            Whether to merge all vertex indices, by default False.

        Returns
        -------
        truth_indices : np.ndarray
            Cleaned truth vertex indices for the tagger.
        reco_indices : np.ndarray
            Cleaned reco vertex indices for the tagger.
        """
        if "vertexing" not in self.aux_tasks:
            raise ValueError("Vertexing aux task not available for this tagger.")
        if "vertexing" not in self.aux_labels:
            raise ValueError("Vertexing labels not found.")
        if "track_origin" not in self.aux_labels:
            raise ValueError("Track origin labels not found.")

        truth_indices = np.copy(self.aux_labels["vertexing"])
        reco_indices = np.copy(self.aux_scores["vertexing"])

        # clean truth and reco indices for each jet
        for i in range(truth_indices.shape[0]):
            truth_indices[i] = clean_truth_vertices(
                truth_indices[i], self.aux_labels["track_origin"][i], incl_vertexing=incl_vertexing
            )
            reco_track_origin = (
                None
                if "track_origin" not in self.aux_scores
                else self.aux_scores["track_origin"][i]
            )
            reco_indices[i] = clean_reco_vertices(
                reco_indices[i], reco_track_origin, incl_vertexing=incl_vertexing
            )

        return truth_indices, reco_indices



"""Auxiliary task results module for high level API."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from ftag import Cuts, Flavours, Label
from ftag.hdf5 import H5Reader

from puma import Histogram, HistogramPlot
from puma.hlplots.tagger import Tagger
from puma.matshow import MatshowPlot
from puma.utils import logger
from puma.utils.aux import get_aux_labels, get_trackOrigin_classNames
from puma.utils.confusion_matrix import confusion_matrix
from puma.utils.mass import calculate_vertex_mass
from puma.utils.precision_recall_scores import precision_recall_scores_per_class
from puma.utils.vertexing import calculate_vertex_metrics
from puma.var_vs_vtx import VarVsVtx, VarVsVtxPlot


@dataclass
class AuxResults:
    """Store information about several taggers and plot auxiliary task results."""

    sample: str
    atlas_first_tag: str = "Simulation Internal"
    atlas_second_tag: str = None
    atlas_third_tag: str = None
    taggers: dict = field(default_factory=dict)
    perf_vars: str | tuple | list = "pt"
    aux_perf_vars: str | tuple | list = None
    output_dir: str | Path = "."
    extension: str = "png"
    global_cuts: Cuts | list | None = None
    num_jets: int | None = None
    remove_nan: bool = False

    def __post_init__(self):
        if isinstance(self.output_dir, str):
            self.output_dir = Path(self.output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        if isinstance(self.perf_vars, str):
            self.perf_vars = [self.perf_vars]
        if self.atlas_second_tag is not None and self.atlas_third_tag is not None:
            self.atlas_second_tag = f"{self.atlas_second_tag}\n{self.atlas_third_tag}"

        self.plot_funcs = {
            "vertexing": self.plot_var_vtx_perf,
            "track_origin": self.plot_track_origin_confmat,
        }

    def add(self, tagger):
        """Add tagger to class.

        Parameters
        ----------
        tagger : puma.hlplots.Tagger
            Instance of the puma.hlplots.Tagger class, containing tagger information.

        Raises
        ------
        KeyError
            if model name duplicated
        """
        if str(tagger) in self.taggers:
            raise KeyError(f"{tagger} was already added.")
        self.taggers[str(tagger)] = tagger

    def load_taggers_from_file(  # pylint: disable=R0913
        self,
        taggers: list[Tagger],
        file_path: Path | str,
        key="jets",
        label_var="HadronConeExclTruthLabelID",
        aux_key="tracks",
        cuts: Cuts | list | None = None,
        num_jets: int | None = None,
        perf_vars: dict | None = None,
    ):
        """Load one or more taggers from a common file.

        Parameters
        ----------
        taggers : list[Tagger]
            List of taggers to add
        file_path : str | Path
            Path to file
        key : str, optional
            Key in file, by default 'jets'
        label_var : str
            Label variable to use, by default 'HadronConeExclTruthLabelID'
        aux_key : str, optional
            Key for auxiliary information, by default 'tracks'
        cuts : Cuts | list, optional
            Cuts to apply, by default None
        num_jets : int, optional
            Number of jets to load from the file, by default all jets
        perf_vars : dict, optional
            Override the performance variable to use, by default None
        """

        def check_nan(data: np.ndarray) -> np.ndarray:
            """
            Filter out NaN values from loaded data.

            Parameters
            ----------
            data : ndarray
                Data to filter
            """
            mask = np.ones(len(data), dtype=bool)
            for name in data.dtype.names:
                mask = np.logical_and(mask, ~np.isnan(data[name]))
            if np.sum(~mask) > 0:
                if self.remove_nan:
                    logger.warning(
                        f"{np.sum(~mask)} NaN values found in loaded data. Removing" " them."
                    )
                    return data[mask]
                raise ValueError(f"{np.sum(~mask)} NaN values found in loaded data.")
            return data

        # set tagger output nodes
        for tagger in taggers:
            if tagger not in self.taggers.values():
                self.add(tagger)

        # get a list of all variables to be loaded from the file
        if not isinstance(cuts, Cuts):
            cuts = Cuts.empty() if cuts is None else Cuts.from_list(cuts)
        var_list = [label_var]
        var_list += cuts.variables
        var_list += sum([t.cuts.variables for t in taggers if t.cuts is not None], [])
        var_list = list(set(var_list + self.perf_vars))

        aux_labels = get_aux_labels()

        aux_var_list = sum([list(t.aux_variables.values()) for t in taggers], [])
        aux_var_list += sum([list(aux_labels.values()) for t in taggers], [])
        if self.aux_perf_vars is not None:
            aux_var_list += self.aux_perf_vars
        aux_var_list = list(set(aux_var_list))

        # load data
        reader = H5Reader(file_path, precision="full", shuffle=False)
        data = reader.load({key: var_list}, num_jets)[key]
        aux_reader = H5Reader(file_path, precision="full", jets_name=aux_key, shuffle=False)
        aux_data = aux_reader.load({aux_key: aux_var_list}, num_jets)[aux_key]

        # check for nan values
        data = check_nan(data)
        # apply common cuts
        if cuts:
            idx, data = cuts(data)
            aux_data = aux_data[idx]
            if perf_vars is not None:
                for name, array in perf_vars.items():
                    perf_vars[name] = array[idx]

        # for each tagger
        for tagger in taggers:
            sel_data = data
            sel_aux_data = aux_data
            sel_perf_vars = perf_vars

            # apply tagger specific cuts
            if tagger.cuts:
                idx, sel_data = tagger.cuts(data)
                sel_aux_data = aux_data[idx]
                if perf_vars is not None:
                    for name, array in sel_perf_vars.items():
                        sel_perf_vars[name] = array[idx]

            # attach data to tagger objects
            tagger.labels = np.array(sel_data[label_var], dtype=[(label_var, "i4")])
            for task in tagger.aux_tasks:
                tagger.aux_scores[task] = sel_aux_data[tagger.aux_variables[task]]
            for task in aux_labels:
                tagger.aux_labels[task] = sel_aux_data[aux_labels[task]]
            if perf_vars is None:
                tagger.perf_vars = {}
                for perf_var in self.perf_vars:
                    if any(x in perf_var for x in ["pt", "mass"]):
                        tagger.perf_vars[perf_var] = sel_data[perf_var] * 0.001
                    else:
                        tagger.perf_vars[perf_var] = sel_data[perf_var]
            else:
                tagger.perf_vars = sel_perf_vars
            tagger.aux_perf_vars = {}
            if self.aux_perf_vars is not None:
                for aux_perf_var in self.aux_perf_vars:
                    if any(x in aux_perf_var for x in ["pt"]):
                        tagger.aux_perf_vars[aux_perf_var] = sel_aux_data[aux_perf_var] * 0.001
                    elif "deta" in aux_perf_var and "eta" in self.perf_vars:
                        tagger.aux_perf_vars["eta"] = sel_aux_data["deta"] + sel_data[
                            "eta"
                        ].reshape(-1, 1)
                    else:
                        tagger.aux_perf_vars[aux_perf_var] = sel_aux_data[aux_perf_var]

    def __getitem__(self, tagger_name: str):
        """Retrieve Tagger object.

        Parameters
        ----------
        tagger_name : str
            Name of model

        Returns
        -------
        Tagger
            Instance of the puma.hlplots.Tagger class, containing tagger information.
        """
        return self.taggers[tagger_name]

    def get_filename(self, plot_name: str, suffix: str | None = None):
        """Get output name.

        Parameters
        ----------
        plot_name : str
            plot name
        suffix : str, optional
            suffix to add to output name, by default None

        Returns
        -------
        str
            output name
        """
        base = f"{self.sample}_{plot_name}"
        if suffix is not None:
            base += f"_{suffix}"
        return Path(self.output_dir / base).with_suffix(f".{self.extension}")

    def plot_var_vtx_perf(
        self,
        vtx_flavours: list[Label] | list[str] | None = None,
        no_vtx_flavours: list[Label] | list[str] | None = None,
        suffix: str | None = None,
        xlabel: str = r"$p_{T}$ [GeV]",
        perf_var: str = "pt",
        incl_vertexing: bool = False,
        vertex_match_requirement: dict | None = None,
        **kwargs,
    ):
        if vtx_flavours is None and no_vtx_flavours is None:
            raise ValueError(
                "Need to specify either vtx_flavours or no_vtx_flavours (or both) to make plots."
            )
        if vtx_flavours is None:
            vtx_flavours = []
        elif no_vtx_flavours is None:
            no_vtx_flavours = []

        if incl_vertexing:
            suffix = "incl" if not suffix else f"{suffix}_incl"

        if vertex_match_requirement is None:
            vertex_match_requirement = {"eff_req": 0.65, "purity_req": 0.5}
        eff_req = round(vertex_match_requirement["eff_req"] * 100, 1)
        purity_req = round(vertex_match_requirement["purity_req"] * 100, 1)
        vtx_match_str = rf"Recall $\geq {eff_req}\%$, Purity $\geq {purity_req}\%$"
        vtx_string = "\nInclusive vertexing" if incl_vertexing else "\nExclusive vertexing"
        atlas_second_tag = self.atlas_second_tag if self.atlas_second_tag else ""
        atlas_second_tag += vtx_string

        # calculate vertexing information for each tagger
        vtx_metrics = {}
        for tagger in self.taggers.values():
            if "vertexing" not in tagger.aux_tasks:
                logger.warning(f"{tagger.name} does not have vertexing aux task defined. Skipping.")
            assert perf_var in tagger.perf_vars, f"{perf_var} not in tagger {tagger.name} data!"

            # get cleaned vertex indices and calculate vertexing metrics
            truth_indices, reco_indices = tagger.vertex_indices(incl_vertexing=incl_vertexing)
            vtx_metrics[tagger.name] = calculate_vertex_metrics(
                reco_indices, truth_indices, **vertex_match_requirement
            )

        if not vtx_metrics:
            raise ValueError("No taggers with vertexing aux task added.")

        # make plots for flavours where vertices are expected
        for flavour in vtx_flavours:
            if isinstance(flavour, str):
                flav = Flavours[flavour]

            # $n_{vtx}^{match}/n_{vtx}^{true}$
            plot_vtx_eff = VarVsVtxPlot(
                mode="efficiency",
                ylabel="Efficiency",
                xlabel=xlabel,
                logy=False,
                atlas_first_tag=self.atlas_first_tag,
                atlas_second_tag=atlas_second_tag + f", {flav.label}\n{vtx_match_str}",
                y_scale=1.4,
            )
            # $n_{vtx}^{match}/n_{vtx}^{reco}$
            plot_vtx_purity = VarVsVtxPlot(
                mode="purity",
                ylabel="Purity",
                xlabel=xlabel,
                logy=False,
                atlas_first_tag=self.atlas_first_tag,
                atlas_second_tag=atlas_second_tag + f", {flav.label}\n{vtx_match_str}",
                y_scale=1.4,
            )
            # $n_{trk}^{match}/n_{trk}^{true}$
            plot_vtx_trk_eff = VarVsVtxPlot(
                mode="efficiency",
                ylabel="Track Assignment Efficiency",
                xlabel=xlabel,
                logy=False,
                atlas_first_tag=self.atlas_first_tag,
                atlas_second_tag=atlas_second_tag + f", {flav.label}\n{vtx_match_str}",
                y_scale=1.4,
            )
            # $n_{trk}^{match}/n_{trk}^{reco}$
            plot_vtx_trk_purity = VarVsVtxPlot(
                mode="purity",
                ylabel="Track Assignment Purity",
                xlabel=xlabel,
                logy=False,
                atlas_first_tag=self.atlas_first_tag,
                atlas_second_tag=atlas_second_tag + f", {flav.label}\n{vtx_match_str}",
                y_scale=1.4,
            )

            for tagger in self.taggers.values():
                if tagger.name not in vtx_metrics:
                    continue
                is_flavour = tagger.is_flav(flav)
                include_sum = vtx_metrics[tagger.name]["track_overlap"][is_flavour] >= 0

                vtx_perf = VarVsVtx(
                    x_var=tagger.perf_vars[perf_var][is_flavour],
                    n_match=vtx_metrics[tagger.name]["n_match"][is_flavour],
                    n_true=vtx_metrics[tagger.name]["n_ref"][is_flavour],
                    n_reco=vtx_metrics[tagger.name]["n_test"][is_flavour],
                    label=tagger.label,
                    colour=tagger.colour,
                    **kwargs,
                )
                vtx_trk_perf = VarVsVtx(
                    x_var=tagger.perf_vars[perf_var][is_flavour],
                    n_match=np.sum(
                        vtx_metrics[tagger.name]["track_overlap"][is_flavour],
                        axis=1,
                        where=include_sum,
                    ),
                    n_true=np.sum(
                        vtx_metrics[tagger.name]["ref_vertex_size"][is_flavour],
                        axis=1,
                        where=include_sum,
                    ),
                    n_reco=np.sum(
                        vtx_metrics[tagger.name]["test_vertex_size"][is_flavour],
                        axis=1,
                        where=include_sum,
                    ),
                    label=tagger.label,
                    colour=tagger.colour,
                    **kwargs,
                )

                plot_vtx_eff.add(vtx_perf, reference=tagger.reference)
                plot_vtx_purity.add(vtx_perf, reference=tagger.reference)
                plot_vtx_trk_eff.add(vtx_trk_perf, reference=tagger.reference)
                plot_vtx_trk_purity.add(vtx_trk_perf, reference=tagger.reference)

            plot_vtx_eff.draw()
            plot_vtx_eff.savefig(self.get_filename(f"{flav}_vtx_eff_vs_{perf_var}", suffix))

            plot_vtx_purity.draw()
            plot_vtx_purity.savefig(self.get_filename(f"{flav}_vtx_purity_vs_{perf_var}", suffix))

            plot_vtx_trk_eff.draw()
            plot_vtx_trk_eff.savefig(self.get_filename(f"{flav}_vtx_trk_eff_vs_{perf_var}", suffix))

            plot_vtx_trk_purity.draw()
            plot_vtx_trk_purity.savefig(
                self.get_filename(f"{flav}_vtx_trk_purity_vs_{perf_var}", suffix)
            )

        # make plots for flavours where vertices are not expected
        for flavour in no_vtx_flavours:
            if isinstance(flavour, str):
                flav = Flavours[flavour]

            plot_vtx_fakes = VarVsVtxPlot(
                mode="fakes",
                ylabel="Fake Rate",
                xlabel=xlabel,
                logy=False,
                atlas_first_tag=self.atlas_first_tag,
                atlas_second_tag=atlas_second_tag + f", {flav.label}",
                y_scale=1.4,
            )

            for tagger in self.taggers.values():
                if tagger.name not in vtx_metrics:
                    continue
                is_flavour = tagger.is_flav(flav)

                vtx_perf = VarVsVtx(
                    x_var=tagger.perf_vars[perf_var][is_flavour],
                    n_match=vtx_metrics[tagger.name]["n_match"][is_flavour],
                    n_true=vtx_metrics[tagger.name]["n_ref"][is_flavour],
                    n_reco=vtx_metrics[tagger.name]["n_test"][is_flavour],
                    label=tagger.label,
                    colour=tagger.colour,
                    **kwargs,
                )

                plot_vtx_fakes.add(vtx_perf, reference=tagger.reference)

            plot_vtx_fakes.draw()
            plot_vtx_fakes.savefig(self.get_filename(f"{flav}_vtx_fakes_vs_{perf_var}", suffix))

    def plot_vertex_mass(
        self,
        vtx_flavours: list[Label] | list[str],
        incl_vertexing: bool = True,
        mass_range: tuple = (0, 5),
        **kwargs,
    ):
        """Plot vertex mass for each tagger.

        Parameters
        ----------
        vtx_flavours : list[Label] | list[str]
            List of jet flavours to make SV mass plots for
        incl_vertexing : bool, optional
            Whether to use inclusive or exclusive vertexing, by default inclusive
        mass_range: tuple, optional
            Range of the mass histogram, by default (0, 5)
        kwargs : dict
            Keyword arguments for `puma.HistogramPlot`
        """
        if incl_vertexing:
            vertexing_text = "Inclusive"
            suffix = "incl"
        else:
            vertexing_text = "Exclusive"
            suffix = "excl"

        atlas_second_tag = "" if self.atlas_second_tag is None else self.atlas_second_tag

        for flavour in vtx_flavours:
            if isinstance(flavour, str):
                flav = Flavours[flavour]

            mass_plot = HistogramPlot(
                bins_range=mass_range,
                xlabel="$m_{SV}$ [GeV]",
                ylabel="Normalized number of vertices",
                atlas_first_tag=self.atlas_first_tag,
                atlas_second_tag=atlas_second_tag + f"\n{vertexing_text} vertexing, {flav.label}",
                y_scale=1.7,
                n_ratio_panels=1,
                **kwargs,
            )

            if incl_vertexing:
                mass_diff_plot = HistogramPlot(
                    bins=np.linspace(-mass_range[1] / 2, mass_range[1] / 2, 12),
                    xlabel=r"$\Delta m_{SV}$ [GeV] (reco - truth)",
                    ylabel="Normalized number of vertices",
                    atlas_first_tag=self.atlas_first_tag,
                    atlas_second_tag=atlas_second_tag
                    + f"\n{vertexing_text} vertexing, {flav.label}",
                    y_scale=1.4,
                    **kwargs,
                )

            for i, tagger in enumerate(self.taggers.values()):
                if "vertexing" not in tagger.aux_tasks:
                    logger.warning(
                        f"{tagger.name} does not have vertexing aux task defined. Skipping."
                    )

                assert {"pt", "eta", "dphi"}.issubset(
                    set(tagger.aux_perf_vars.keys())
                ), "Track pt, eta or dphi not in tagger.aux_perf_vars (required to calculate \
                    vertex masses). Track eta is automatically calculated if deta and jet eta \
                    are supplied."

                # get cleaned vertex indices and calculate vertexing metrics
                truth_indices, reco_indices = tagger.vertex_indices(incl_vertexing=incl_vertexing)
                is_flavour = tagger.is_flav(flav)

                if i == 0:
                    masses = calculate_vertex_mass(
                        tagger.aux_perf_vars["pt"][is_flavour],
                        tagger.aux_perf_vars["eta"][is_flavour],
                        tagger.aux_perf_vars["dphi"][is_flavour],
                        truth_indices[is_flavour],
                        particle_mass=0.13957,  # pion mass in GeV
                    )

                    if incl_vertexing:
                        truth_masses = np.max(masses, axis=1)
                    else:
                        truth_masses = np.unique(masses, axis=1).flatten()

                    mass_plot.add(
                        Histogram(
                            truth_masses[truth_masses > 0.14],
                            label="MC truth",
                            colour="#000000",
                        ),
                        reference=True,
                    )

                masses = calculate_vertex_mass(
                    tagger.aux_perf_vars["pt"][is_flavour],
                    tagger.aux_perf_vars["eta"][is_flavour],
                    tagger.aux_perf_vars["dphi"][is_flavour],
                    reco_indices[is_flavour],
                    particle_mass=0.13957,  # pion mass in GeV
                )

                if incl_vertexing:
                    sv_masses = np.max(masses, axis=1)
                    mass_diffs = sv_masses - truth_masses
                    mass_diffs = mass_diffs[np.logical_and(truth_masses > 0.14, sv_masses > 0.14)]
                    mass_diff_plot.add(
                        Histogram(mass_diffs, label=tagger.label, colour=tagger.colour)
                    )
                else:
                    sv_masses = np.concatenate([np.unique(imass) for imass in masses])

                sv_masses = sv_masses[sv_masses > 0.14]  # remove single and zero track vertices
                mass_plot.add(Histogram(sv_masses, label=tagger.label, colour=tagger.colour))

            mass_plot.draw()
            mass_plot.savefig(self.get_filename(f"{flav}_sv_mass_{suffix}"))

            if incl_vertexing:
                mass_diff_plot.draw()
                mass_diff_plot.savefig(self.get_filename(f"{flav}_sv_mass_diff"))

    def plot_track_origin_confmat(
        self,
        normalize: str | None = "rownorm",
        minimal_plot: bool = True,
        **kwargs,
    ):
        """Plot Track Origin Aux Task confusion matrix.

        Parameters
        ----------
        normalize : str | None, optional
            Normalization of the confusion matrix. Can be:
            None: Give raw counts;
            "rownorm": Normalize across the prediction class, i.e. such that the rows add to one;
            "colnorm": Normalize across the target class, i.e. such that the columns add to one;
            "all" : Normalize across all examples, i.e. such that all matrix entries add to one.
            Defaults to "rownorm".
        minimal_plot : bool, optional
            Whether to plot the CM with minimal or full info (title, cbar), by default True
        **kwargs : kwargs
            Keyword arguments for `puma.MatshowPlot` and `puma.PlotObject`
        """
        for tagger in self.taggers.values():
            # Reading tagger's target and predicted labels
            # and flattening them so that they have shape (Ntracks,)
            target = tagger.aux_labels["track_origin"].reshape(-1)
            predictions = tagger.aux_scores["track_origin"].reshape(-1)

            padding_removal = target >= 0

            target = target[padding_removal]
            predictions = predictions[padding_removal]

            # Computing the confusion matrix
            cm = confusion_matrix(target, predictions, normalize=normalize)
            precision, recall = precision_recall_scores_per_class(target, predictions)

            class_names = get_trackOrigin_classNames()
            class_names_with_perf = []

            if minimal_plot:
                for i, c in enumerate(class_names):
                    # class_names_with_perf.append(f"{c}\nRecall = {recall[i]:.3f}")
                    class_names_with_perf.append(
                        f"{c}\nPrecision = {precision[i]:.3f}\nRecall = {recall[i]:.3f}"
                    )
                # Plotting the confusion matrix
                plot_cm = MatshowPlot(
                    x_ticklabels=class_names,
                    y_ticklabels=class_names_with_perf,
                    xlabel="Predicted Classes",
                    ylabel="Target Classes",
                    atlas_second_tag=self.atlas_second_tag,
                    show_cbar=False,
                    atlas_tag_outside=True,
                    **kwargs,
                )
            else:
                for i, c in enumerate(class_names):
                    class_names_with_perf.append(
                        f"{c}\nPrecision = {precision[i]:.3f}\nRecall = {recall[i]:.3f}"
                    )
                # Plotting the confusion matrix
                plot_cm = MatshowPlot(
                    x_ticklabels=class_names,
                    y_ticklabels=class_names_with_perf,
                    title="Track Origin Auxiliary Task\nConfusion Matrix",
                    xlabel="Predicted Classes",
                    ylabel="Target Classes",
                    atlas_second_tag=self.atlas_second_tag,
                    atlas_tag_outside=True,
                    show_cbar=False,
                    **kwargs,
                )
            plot_cm.draw(cm)
            base = tagger.name + "_trackOrigin_cm"
            plot_cm.savefig(self.get_filename(base))



"""Unit test script for the functions in var_vs_var.py."""

from __future__ import annotations

import os
import tempfile
import unittest

import numpy as np
from matplotlib.testing.compare import compare_images

# from matplotlib.testing.compare import compare_images
from puma import VarVsVtx, VarVsVtxPlot
from puma.utils.logging import logger, set_log_level

set_log_level(logger, "DEBUG")


class VarVsVtxTestCase(unittest.TestCase):
    """Test class for the puma.var_vs_vtx functions."""

    def setUp(self):
        np.random.seed(42)

        self.x_var = np.linspace(100, 250, 20)
        self.n_true = np.random.randint(1, 10, size=20)
        self.n_reco = np.random.randint(1, 10, size=20)
        self.n_match = np.random.randint(0, np.minimum(self.n_true, self.n_reco))

    def test_var_vs_vtx_init_wrong_true_shape(self):
        """Test var_vs_vtx init."""
        with self.assertRaises(ValueError):
            VarVsVtx(
                x_var=np.ones(4),
                n_match=np.ones(4),
                n_true=np.ones(5),
                n_reco=np.ones(4),
            )

    def test_var_vs_vtx_init_wrong_reco_shape(self):
        """Test var_vs_vtx init."""
        with self.assertRaises(ValueError):
            VarVsVtx(
                x_var=np.ones(4),
                n_match=np.ones(4),
                n_true=np.ones(4),
                n_reco=np.ones(5),
            )

    def test_var_vs_vtx_init_wrong_match_shape(self):
        """Test var_vs_vtx init."""
        with self.assertRaises(ValueError):
            VarVsVtx(
                x_var=np.ones(4),
                n_match=np.ones(5),
                n_true=np.ones(4),
                n_reco=np.ones(4),
            )

    def test_var_vs_vtx_init(self):
        """Test var_vs_vtx init."""
        VarVsVtx(
            x_var=np.ones(6),
            n_match=np.ones(6),
            n_true=np.ones(6),
            n_reco=np.ones(6),
            bins=10,
            key="test",
        )

    def test_var_vs_vtx_set_bin_edges_list(self):
        """Test var_vs_vtx set_bin_edges."""
        var_plot = VarVsVtx(
            x_var=[0, 1, 2],
            n_match=[3, 4, 5],
            n_true=[3, 4, 5],
            n_reco=[3, 4, 5],
            bins=[0, 1, 2],
        )
        np.testing.assert_array_almost_equal(var_plot.bin_edges, [0, 1, 2])

    def test_var_vs_vtx_set_bin_edges(self):
        """Test var_vs_vtx set_bin_edges."""
        var_plot = VarVsVtx(
            x_var=[0, 1, 2],
            n_match=[3, 4, 5],
            n_true=[3, 4, 5],
            n_reco=[3, 4, 5],
            bins=2,
        )
        np.testing.assert_array_almost_equal(var_plot.bin_edges, [0, 1, 2], decimal=4)

    def test_var_vs_vtx_get_perf_ratio_zero(self):
        """Test var_vs_vtx get_performance_ratio with zero efficiency case."""
        var_plot = VarVsVtx(
            x_var=[0, 1, 2],
            n_match=[0, 0, 0],
            n_true=[3, 4, 5],
            n_reco=[3, 4, 5],
            bins=2,
        )
        pm, pm_error = var_plot.get_performance_ratio(var_plot.n_match, var_plot.n_true)
        self.assertEqual(pm, 0)
        self.assertEqual(pm_error, 0)

    def test_var_vs_vtx_get_perf_ratio_infinity(self):
        """Test var_vs_vtx get_performance_ratio with undefined efficiency case."""
        var_plot = VarVsVtx(
            x_var=[0, 1, 2],
            n_match=[0, 0, 0],
            n_true=[0, 0, 0],
            n_reco=[0, 0, 0],
            bins=2,
        )
        pm, pm_error = var_plot.get_performance_ratio(var_plot.n_match, var_plot.n_true)
        np.testing.assert_equal(pm, np.nan)
        np.testing.assert_equal(pm_error, np.nan)

    def test_var_vs_vtx_eq_different_classes(self):
        """Test var_vs_vtx eq."""
        var_plot = VarVsVtx(
            x_var=[0, 1, 2],
            n_match=[3, 4, 5],
            n_true=[3, 4, 5],
            n_reco=[3, 4, 5],
            bins=2,
        )
        self.assertNotEqual(var_plot, np.ones(6))

    def test_var_vs_vtx_get(self):
        var_plot = VarVsVtx(
            x_var=self.x_var,
            n_match=self.n_match,
            n_true=self.n_true,
            n_reco=self.n_reco,
            bins=1,
        )
        mode_options = ["efficiency", "purity", "fakes"]
        for mode in mode_options:
            var_plot.get(mode)


class VarVsVtxPlotTestCase(unittest.TestCase):
    """Test class for the puma.var_vs_vtx_plot."""

    def setUp(self):
        # Set up temp directory for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint:disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")
        np.random.seed(42)
        n_random = 10_000

        # truth variables - shared between taggers
        self.x_var = np.random.uniform(0, 250, size=n_random)
        self.n_true = np.random.randint(1, 10, size=n_random)

        # tagger 1
        self.n_reco_1 = np.random.randint(1, 10, size=n_random)
        self.n_match_1 = np.random.randint(0, np.minimum(self.n_true, self.n_reco_1))

        # tagger 2
        self.n_reco_2 = np.random.randint(1, 10, size=n_random)
        self.n_match_2 = np.random.randint(0, np.minimum(self.n_true, self.n_reco_2))

        # Define pT bins
        self.bins = [20, 30, 40, 60, 85, 110, 140, 175, 250]

    def test_var_vs_vtx_plot_mode_wrong_option(self):
        with self.assertRaises(ValueError):
            VarVsVtxPlot(
                mode="test",
                ylabel="Vtx purity",
                xlabel=r"$p_{T}$ [GeV]",
                logy=True,
                atlas_second_tag="test",
                y_scale=1.5,
                n_ratio_panels=1,
                figsize=(9, 6),
            )

    def test_var_vs_vtx_plot_mode_efficiency(self):
        """Test vtx output plot - efficiency."""
        # define the curves
        tagger1 = VarVsVtx(
            x_var=self.x_var,
            n_true=self.n_true,
            n_reco=self.n_reco_1,
            n_match=self.n_match_1,
            bins=self.bins,
            label="tagger 1",
        )
        tagger2 = VarVsVtx(
            x_var=self.x_var,
            n_true=self.n_true,
            n_reco=self.n_reco_2,
            n_match=self.n_match_2,
            bins=self.bins,
            label="tagger 2",
        )

        plot_eff = VarVsVtxPlot(
            mode="efficiency",
            ylabel="Vtx efficiency",
            xlabel=r"$p_{T}$ [GeV]",
            logy=True,
            y_scale=1.5,
            n_ratio_panels=1,
            figsize=(9, 6),
        )
        plot_eff.add(tagger1, reference=True)
        plot_eff.add(tagger2)

        plot_eff.draw()

        plotname = "test_vtx_efficiency.png"
        plot_eff.savefig(f"{self.actual_plots_dir}/{plotname}")
        # Uncomment line below to update expected image
        # plot_eff.savefig(f"{self.expected_plots_dir}/{plotname}")

        # Investigate small shifts in labels/scale causing this test to fail
        self.assertEqual(
            None,
            compare_images(
                f"{self.actual_plots_dir}/{plotname}",
                f"{self.expected_plots_dir}/{plotname}",
                tol=5,
            ),
        )

    def test_var_vs_vtx_plot_mode_purity(self):
        """Test vtx output plot - purity."""
        # define the curves
        tagger1 = VarVsVtx(
            x_var=self.x_var,
            n_true=self.n_true,
            n_reco=self.n_reco_1,
            n_match=self.n_match_1,
            bins=self.bins,
            label="tagger 1",
        )
        tagger2 = VarVsVtx(
            x_var=self.x_var,
            n_true=self.n_true,
            n_reco=self.n_reco_2,
            n_match=self.n_match_2,
            bins=self.bins,
            label="tagger 2",
        )

        plot_pur = VarVsVtxPlot(
            mode="purity",
            ylabel="Vtx purity",
            xlabel=r"$p_{T}$ [GeV]",
            logy=True,
            y_scale=1.5,
            n_ratio_panels=1,
            figsize=(9, 6),
        )
        plot_pur.add(tagger1, reference=True)
        plot_pur.add(tagger2)

        plot_pur.draw()

        plotname = "test_vtx_purity.png"
        plot_pur.savefig(f"{self.actual_plots_dir}/{plotname}")
        # Uncomment line below to update expected image
        # plot_pur.savefig(f"{self.expected_plots_dir}/{plotname}")

        # Investigate small shifts in labels/scale causing this test to fail
        self.assertEqual(
            None,
            compare_images(
                f"{self.actual_plots_dir}/{plotname}",
                f"{self.expected_plots_dir}/{plotname}",
                tol=5,
            ),
        )

    def test_var_vs_vtx_plot_mode_fake_rate(self):
        """Test aux output plot - fake_rate."""
        # define the curves
        tagger1 = VarVsVtx(
            x_var=self.x_var,
            n_true=self.n_true,
            n_reco=self.n_reco_1,
            n_match=self.n_match_1,
            bins=self.bins,
            label="tagger 1",
        )
        tagger2 = VarVsVtx(
            x_var=self.x_var,
            n_true=self.n_true,
            n_reco=self.n_reco_2,
            n_match=self.n_match_2,
            bins=self.bins,
            label="tagger 2",
        )

        plot_fakes = VarVsVtxPlot(
            mode="fakes",
            ylabel="Vtx fake rate",
            xlabel=r"$p_{T}$ [GeV]",
            logy=True,
            y_scale=1.5,
            n_ratio_panels=1,
            figsize=(9, 6),
        )
        plot_fakes.add(tagger1, reference=True)
        plot_fakes.add(tagger2)

        plot_fakes.draw()

        plotname = "test_vtx_fake_rate.png"
        plot_fakes.savefig(f"{self.actual_plots_dir}/{plotname}")
        # Uncomment line below to update expected image
        # plot_fakes.savefig(f"{self.expected_plots_dir}/{plotname}")

        # Investigate small shifts in labels/scale causing this test to fail
        self.assertEqual(
            None,
            compare_images(
                f"{self.actual_plots_dir}/{plotname}",
                f"{self.expected_plots_dir}/{plotname}",
                tol=5,
            ),
        )



"""Unit test script for the functions in hlplots/tagger.py."""

from __future__ import annotations

import tempfile
import unittest
from pathlib import Path

import h5py
import yaml
from ftag import Flavours, get_mock_file
from ftag.hdf5 import structured_from_dict
from yamlinclude import YamlIncludeConstructor

from puma.hlplots.yuma import YumaConfig, main
from puma.hlplots.yutils import get_tagger_name

EXAMPLES = Path(__file__).parents[3] / "examples"


def load_no_include(plt_cfg, taggers):
    def dummy_inc(loader, node):  # noqa: ARG001
        return node.value

    # Don't load the taggers
    yaml.SafeLoader.add_constructor("!include", dummy_inc)

    plt_cfg = EXAMPLES / "plt_cfg.yaml"
    with open(plt_cfg) as f:
        plt_cfg = yaml.safe_load(f)
    taggers = EXAMPLES / "taggers.yaml"
    with open(taggers) as f:
        taggers = yaml.safe_load(f)

    return plt_cfg, taggers


class TestYutils(unittest.TestCase):
    def setUp(self):
        self.flavours = [Flavours[f] for f in ["ujets", "cjets", "bjets"]]
        # support inclusion of yaml files in the config dir
        YamlIncludeConstructor.add_to_loader_class(loader_class=yaml.SafeLoader, base_dir=EXAMPLES)

    def testGetIncludeTaggers(self):
        plt_cfg = EXAMPLES / "plt_cfg.yaml"
        taggers = EXAMPLES / "taggers.yaml"
        plt_cfg, taggers = load_no_include(plt_cfg, taggers)

        with tempfile.TemporaryDirectory() as tmp_file:
            fpath1, _file = get_mock_file(fname=(Path(tmp_file) / "file1.h5").as_posix())
            taggers["dummy1"]["sample_path"] = fpath1
            taggers["dummy2"]["sample_path"] = fpath1
            taggers["dummy3"]["sample_path"] = fpath1
            updated_plt_cfg = Path(tmp_file) / "plt_cfg.yaml"

            plt_cfg["plots"]["roc"][0]["reference"] = "dummyNot"
            plt_cfg["plots"]["roc"][0]["include_taggers"] = ["dummy1"]

            plt_cfg["plot_dir"] = tmp_file + "/plots"
            plt_cfg["taggers_config"] = taggers

            with open(updated_plt_cfg, "w") as f:
                yaml.dump(plt_cfg, f)
            print(plt_cfg)
            args = ["--config", updated_plt_cfg.as_posix(), "--signals", "bjets"]
            main(args)

    def testGetTaggerName(self):
        fpath, _file = get_mock_file()
        name = get_tagger_name(None, fpath, key="TestName1", flavours=self.flavours)
        assert name == "MockTagger"

    def testBreakGetTaggerName(self):
        _fpath, file = get_mock_file()
        updated = {k: file["jets"][k] for k in file["jets"].dtype.names}
        updated["Tagger2_pu"] = updated["MockTagger_pu"]
        updated["Tagger2_pb"] = updated["MockTagger_pb"]
        updated["Tagger2_pc"] = updated["MockTagger_pc"]
        array = structured_from_dict(updated)
        with tempfile.TemporaryDirectory() as tmp_file:
            fname = Path(tmp_file) / "test.h5"
            with h5py.File(fname, "w") as f:
                f.create_dataset("jets", data=array)
            with self.assertRaises(ValueError):
                get_tagger_name(None, fname, key=None, flavours=self.flavours)

    def testGetSignals(self):
        plt_cfg = EXAMPLES / "plt_cfg.yaml"
        plt_cfg = YumaConfig.load_config(plt_cfg)
        assert sorted(plt_cfg.signals) == ["bjets", "cjets"]


class TestYumaPlots(unittest.TestCase):
    def testArgs(self):
        config_path = str(EXAMPLES / "plt_cfg.yaml")
        args = ["--config", config_path, "--signals", "bjets", "--plots", "not_valid"]
        with self.assertRaises(SystemExit):
            main(args)

    def testCheckConfig(self):
        with tempfile.TemporaryDirectory() as tmp_file:
            cpath = Path(tmp_file) / "dummyconfig.yaml"
            tmp_config = {
                "plot_dir": "test/dir",
                "taggers_config": {"tagger1": {"sample_path": "dummy1"}},
                "results_config": {"sample": "ttbar"},
                "plots": {
                    "roc": [
                        {
                            "signal": "bjets",
                            "include_taggers": ["tagger1"],
                            "efficiency": 0.8,  # Shouldn't be here
                        }
                    ]
                },
            }
            with open(cpath, "w") as f:
                yaml.dump(tmp_config, f)
            args = ["--config", cpath.as_posix(), "--signals", "bjets"]
            with self.assertRaises(ValueError):
                main(args)

    def testAllPlots(self):
        plt_cfg = EXAMPLES / "plt_cfg.yaml"
        taggers = EXAMPLES / "taggers.yaml"
        plt_cfg, taggers = load_no_include(plt_cfg, taggers)

        with tempfile.TemporaryDirectory() as tmp_file:
            fpath1, _file = get_mock_file(fname=(Path(tmp_file) / "file1.h5").as_posix())
            fpath2, _file = get_mock_file(fname=(Path(tmp_file) / "file2.h5").as_posix())
            taggers["dummy1"]["sample_path"] = fpath1
            taggers["dummy2"]["sample_path"] = fpath1
            taggers["dummy3"]["sample_path"] = fpath2
            updated_plt_cfg = Path(tmp_file) / "plt_cfg.yaml"
            plt_cfg["plot_dir"] = tmp_file + "/plots"
            plt_cfg["taggers_config"] = taggers

            with open(updated_plt_cfg, "w") as f:
                yaml.dump(plt_cfg, f)

            args = ["--config", updated_plt_cfg.as_posix(), "--signals", "bjets"]
            main(args)

            # Simple check on number of output plots
            out_dir = Path(tmp_file) / "plots" / "plt_cfg"
            btagging = out_dir / "btag"
            ctagging = out_dir / "ctag"
            assert btagging.exists(), "No b-tagging plots produced"
            assert not ctagging.exists(), "No c-tagging plots should have been produced"
            btag_plots = [p.name for p in btagging.rglob("*.png")]
            assert len(btag_plots) == 22, f"Expected 22 b-tagging plot, found {len(btag_plots)}"

            args = [
                "--config",
                updated_plt_cfg.as_posix(),
                "--plots",
                "roc",
                "--signals",
                "cjets",
            ]
            main(args)

            ctag_plots = [p.name for p in ctagging.rglob("*.png")]
            assert ctagging.exists(), "No c-tagging plots produced"
            assert (
                len(ctag_plots) == 1
            ), f"Only expected one c-tagging plot, found {len(ctag_plots)}: , {ctag_plots}"

            args = ["--config", updated_plt_cfg.as_posix()]
            main(args)

    def testNoPlots(self):
        plt_cfg = EXAMPLES / "plt_cfg.yaml"
        taggers = EXAMPLES / "taggers.yaml"

        plt_cfg, taggers = load_no_include(plt_cfg, taggers)

        with tempfile.TemporaryDirectory() as tmp_file:
            fpath1, _file = get_mock_file(fname=(Path(tmp_file) / "file1.h5").as_posix())
            fpath2, _file = get_mock_file(fname=(Path(tmp_file) / "file2.h5").as_posix())

            taggers["dummy1"]["sample_path"] = fpath1
            taggers["dummy2"]["sample_path"] = fpath1
            taggers["dummy3"]["sample_path"] = fpath2

            updated_plt_cfg = Path(tmp_file) / "plt_cfg.yaml"
            plt_cfg["taggers_config"] = taggers
            plt_cfg["plot_dir"] = tmp_file + "/plots"

            plt_cfg["plots"] = {"roc": plt_cfg["plots"]["roc"]}
            with open(updated_plt_cfg, "w") as f:
                yaml.dump(plt_cfg, f)

            args = ["--config", updated_plt_cfg.as_posix(), "--signals", "bjets"]
            main(args)

            # Simple check on number of output plots
            out_dir = Path(tmp_file) / "plots" / "plt_cfg"
            btagging = out_dir / "btag"
            ctagging = out_dir / "ctag"
            assert btagging.exists(), "No b-tagging plots produced"
            assert not ctagging.exists(), "No c-tagging plots should have been produced"
            btag_plots = [p.name for p in btagging.rglob("*.png")]
            print(btag_plots)
            assert len(btag_plots) == 3, f"Expected 3 b-tagging plot, found {len(btag_plots)}"



"""Unit test script for the functions in utils/vertexing.py."""

from __future__ import annotations

import unittest

import numpy as np

from puma.utils import logger, set_log_level
from puma.utils.mass import (
    calculate_vertex_mass,
)

set_log_level(logger, "DEBUG")


class VertexMassTestCase(unittest.TestCase):
    """Test case for clean_indices function."""

    def test_output_shape(self):
        """Check that output has the right shape."""
        track_pt = np.zeros((5, 10))
        track_eta = np.zeros((5, 10))
        track_phi = np.zeros((5, 10))
        vtx_idx = np.zeros((5, 10))
        vtx_masses = calculate_vertex_mass(track_pt, track_eta, track_phi, vtx_idx)
        expected_shape = (5, 10)
        np.testing.assert_array_equal(vtx_masses.shape, expected_shape)

    def test_single_track_only(self):
        """Check case where only single track vertices exist."""
        track_pt = np.zeros((1, 5))
        track_eta = np.zeros((1, 5))
        track_phi = np.zeros((1, 5))
        vtx_idx = np.array([[0, 1, 2, 3, 4]])
        vtx_masses = calculate_vertex_mass(track_pt, track_eta, track_phi, vtx_idx, particle_mass=2)
        expected_result = np.array([[2, 2, 2, 2, 2]])
        np.testing.assert_array_equal(vtx_masses, expected_result)

    def test_multiple_vertices(self):
        """Check more complicated case with two vertices."""
        track_pt = np.random.rand(1, 5) * 1000
        track_eta = np.random.rand(1, 5) * 2
        track_phi = np.random.rand(1, 5) * 2 * np.pi - np.pi
        vtx_idx = np.array([[0, 0, 1, 1, 1]])

        px = track_pt * np.cos(track_phi)
        py = track_pt * np.sin(track_phi)
        pz = track_pt * np.sinh(track_eta)
        e = np.sqrt(px**2 + py**2 + pz**2 + 0.13957**2)

        vtx_mass_1 = np.sqrt(
            (e[0, 0] + e[0, 1]) ** 2
            - (px[0, 0] + px[0, 1]) ** 2
            - (py[0, 0] + py[0, 1]) ** 2
            - (pz[0, 0] + pz[0, 1]) ** 2
        )
        vtx_mass_2 = np.sqrt(
            (e[0, 2] + e[0, 3] + e[0, 4]) ** 2
            - (px[0, 2] + px[0, 3] + px[0, 4]) ** 2
            - (py[0, 2] + py[0, 3] + py[0, 4]) ** 2
            - (pz[0, 2] + pz[0, 3] + pz[0, 4]) ** 2
        )

        vtx_masses = calculate_vertex_mass(
            track_pt, track_eta, track_phi, vtx_idx, particle_mass=0.13957
        )
        expected_result = np.array([[vtx_mass_1, vtx_mass_1, vtx_mass_2, vtx_mass_2, vtx_mass_2]])
        np.testing.assert_array_equal(vtx_masses, expected_result)



"""Aux task efficiency plots vs. specific variable."""

from __future__ import annotations

from typing import ClassVar

import numpy as np

from puma.metrics import eff_err
from puma.utils import logger
from puma.utils.histogram import save_divide
from puma.var_vs_var import VarVsVar, VarVsVarPlot


class VarVsVtx(VarVsVar):  # pylint: disable=too-many-instance-attributes
    """var_vs_vtx class storing info about vertexing performance."""

    def __init__(
        self,
        x_var: np.ndarray,
        n_match: np.ndarray,
        n_true: np.ndarray,
        n_reco: np.ndarray,
        bins=10,
        key: str | None = None,
        **kwargs,
    ) -> None:
        """Initialise properties of roc curve object.

        Parameters
        ----------
        x_var : np.ndarray
            Values for x-axis variable for signal
        n_match : np.ndarray
            Values for number of correctly identified objects (where truth and
            reco match)
        n_true : np.ndarray
            Values for true number of objects
        n_reco : np.ndarray
            Values for reconstructed number of objects
        bins : int or sequence of scalars, optional
            If bins is an int, it defines the number of equal-width bins in the
            given range (10, by default). If bins is a sequence, it defines a
            monotonically increasing array of bin edges, including the
            rightmost edge, allowing for non-uniform bin widths, by default 10
        key : str, optional
            Identifier for the curve e.g. tagger, by default None
        **kwargs : kwargs
            Keyword arguments passed to `PlotLineObject`

        Raises
        ------
        ValueError
            If provided options are not compatible with each other
        """
        if len(x_var) != len(n_match):
            raise ValueError(
                f"Length of `x_var` ({len(x_var)}) and `n_match` "
                f"({len(n_match)}) have to be identical."
            )
        if len(x_var) != len(n_true):
            raise ValueError(
                f"Length of `x_var` ({len(x_var)}) and `n_true` "
                f"({len(n_true)}) have to be identical."
            )
        if len(x_var) != len(n_reco):
            raise ValueError(
                f"Length of `x_var` ({len(x_var)}) and `n_reco` "
                f"({len(n_reco)}) have to be identical."
            )

        self.x_var = np.array(x_var)
        self.n_match = np.array(n_match)
        self.n_true = np.array(n_true)
        self.n_reco = np.array(n_reco)
        # Binning related variables
        self.n_bins = None
        self.bin_edges = None
        self.x_bin_centres = None
        self.bin_widths = None
        # Binned distributions
        self.bin_indices = None
        self.metric_binned = None

        self._set_bin_edges(bins)
        self._apply_binning()

        VarVsVar.__init__(
            self,
            x_var=self.x_bin_centres,
            y_var_mean=np.zeros_like(self.x_bin_centres),
            y_var_std=np.zeros_like(self.x_bin_centres),
            x_var_widths=2 * self.bin_widths,
            key=key,
            fill=True,
            plot_y_std=False,
            **kwargs,
        )

    def _set_bin_edges(self, bins):
        """Calculate bin edges, centres and width and save them as class variables.

        Parameters
        ----------
        bins : int or sequence of scalars
            If bins is an int, it defines the number of equal-width bins in the given
            range. If bins is a sequence, it defines a monotonically increasing array of
            bin edges, including the rightmost edge, allowing for non-uniform bin
            widths.
        """
        logger.debug("Calculating binning.")
        if isinstance(bins, int):
            # With this implementation, the data point with x=xmax will be added to the
            # overflow bin.
            xmin, xmax = np.amin(self.x_var), np.amax(self.x_var)
            # increasing xmax slightly to inlcude largest value due to hehavior of
            # np.digitize
            xmax *= 1 + 1e-5
            self.bin_edges = np.linspace(xmin, xmax, bins + 1)
        elif isinstance(bins, (list, np.ndarray)):
            self.bin_edges = np.array(bins)
        logger.debug("Retrieved bin edges %s}", self.bin_edges)
        # Get the bins for the histogram
        self.x_bin_centres = (self.bin_edges[:-1] + self.bin_edges[1:]) / 2.0
        self.bin_widths = (self.bin_edges[1:] - self.bin_edges[:-1]) / 2.0
        self.n_bins = self.bin_edges.size - 1
        logger.debug("N bins: %i", self.n_bins)

    def _apply_binning(self):
        """Get binned distributions for number of matches, truth and reco objects."""
        logger.debug("Applying binning.")
        self.bin_indices = np.digitize(self.x_var, self.bin_edges)
        self.match_binned = [
            self.n_match[np.where(self.bin_indices == x)[0]] for x in range(1, len(self.bin_edges))
        ]
        self.true_binned = [
            self.n_true[np.where(self.bin_indices == x)[0]] for x in range(1, len(self.bin_edges))
        ]
        self.reco_binned = [
            self.n_reco[np.where(self.bin_indices == x)[0]] for x in range(1, len(self.bin_edges))
        ]

    def get_performance_ratio(self, num: np.ndarray, denom: np.ndarray):
        """Calculate performance ratio for vertexing task. Either n_matched/n_true
        (efficiency) or n_matched/n_reco (purity).

        Parameters
        ----------
        arr : np.ndarray
            Array with discriminants
        cut : float
            Cut value

        Returns
        -------
        float
            Performance ratio
        float
            Performance ratio error
        """
        pm = save_divide(np.sum(num), np.sum(denom), default=np.inf)
        if pm == np.inf:
            logger.warning("Your vertexing performance ratio is infinity -> setting it to np.nan.")
            return np.nan, np.nan
        if pm == 0:
            logger.warning("Your vertexing performance ratio is zero -> setting error to zero.")
            return 0.0, 0.0
        pm_error = eff_err(pm, len(num))
        return pm, pm_error

    @property
    def efficiency(self):
        """Calculate vertexing efficiency per bin. Defined as number of reconstructed
        vertices matched to truth divided by number of total true vertices.

        Returns
        -------
        np.ndarray
            Efficiency
        np.ndarray
            Efficiency error
        """
        logger.debug("Calculating vertexing efficiency.")
        eff = list(
            map(
                self.get_performance_ratio,
                self.match_binned,
                self.true_binned,
            )
        )
        logger.debug("Retrieved vertexing efficiencies: %s", eff)
        return np.array(eff)[:, 0], np.array(eff)[:, 1]

    @property
    def purity(self):
        """Calculate vertexing purity per bin. Defined as number of reconstructed
        vertices matched to truth divided by number of total reconstructed vertices.

        Returns
        -------
        np.ndarray
            Purity
        np.ndarray
            Purity error
        """
        logger.debug("Calculating vertexing purity.")
        purity = list(
            map(
                self.get_performance_ratio,
                self.match_binned,
                self.reco_binned,
            )
        )
        logger.debug("Retrieved vertexing purity: %s", purity)
        return np.array(purity)[:, 0], np.array(purity)[:, 1]

    @property
    def fakes(self):
        """Calculate vertexing fake rate per bin. Defined as total number
        of events with reconstructed vertices where vertices are not expected.

        Returns
        -------
        np.ndarray
            Fake rate
        np.ndarray
            Fake rate error
        """
        logger.debug("Calculating vertexing fake rate.")
        total_reco = list(
            map(
                self.get_performance_ratio,
                [np.where(reco_bin > 0, 1, 0) for reco_bin in self.reco_binned],
                list(map(np.ones_like, self.reco_binned)),
            )
        )
        logger.debug("Retrieved vertexing fake rate: %s", total_reco)
        return np.array(total_reco)[:, 0], np.array(total_reco)[:, 1]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                np.all(self.x_var == other.x_var)
                and np.all(self.n_match == other.n_match)
                and np.all(self.n_true == other.n_true)
                and np.all(self.n_reco == other.n_reco)
                and np.all(self.bin_edges == other.bin_edges)
                and self.key == other.key
            )
        return False

    def get(self, mode: str):
        """Wrapper around rejection and efficiency functions.

        Parameters
        ----------
        mode : str
            Can be "efficiency", "purity" or "fakes"

        Returns
        -------
        np.ndarray
            Efficiency, purity or fake rate depending on `mode` value
        np.ndarray
            Efficiency, purity or fake rate error depending on `mode` value

        Raises
        ------
        ValueError
            If mode not supported
        """
        if mode == "efficiency":
            return self.efficiency
        if mode == "purity":
            return self.purity
        if mode == "fakes":
            return self.fakes
        raise ValueError(
            f"The selected mode {mode} is not supported. Use one of the following:"
            f" {VarVsVtxPlot.mode_options}."
        )


class VarVsVtxPlot(VarVsVarPlot):  # pylint: disable=too-many-instance-attributes
    mode_options: ClassVar[list[str]] = [
        "efficiency",
        "purity",
        "fakes",
    ]

    def __init__(self, mode, grid: bool = False, **kwargs) -> None:
        """var_vs_vtx plot properties.

        Parameters
        ----------
        mode : str
            Defines which quantity is plotted, the following options ar available:
                efficiency - Plots efficiency vs. variable for jets where vertices are
                expected
                purity - Plots purity vs. variable for jets where vertices are expected
                fakes - Plots fake rate vs. variable for jets where vertices are not
                expected
        grid : bool, optional
            Set the grid for the plots.
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`

        Raises
        ------
        ValueError
            If incompatible mode given or more than 1 ratio panel requested
        """
        super().__init__(grid=grid, **kwargs)
        if mode not in self.mode_options:
            raise ValueError(
                f"The selected mode {mode} is not supported. Use one of the following: "
                f"{self.mode_options}."
            )
        self.mode = mode

    def _setup_curves(self):
        for key in self.add_order:
            elem = self.plot_objects[key]
            y_value, y_error = elem.get(self.mode)
            elem.y_var_mean = y_value
            elem.y_var_std = y_error

    def plot(self, **kwargs):
        """Plotting curves.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to plt.axis.errorbar

        Returns
        -------
        Line2D
            matplotlib Line2D object
        """
        logger.debug("Plotting curves with mode %s", self.mode)
        self._setup_curves()
        return super().plot(**kwargs)



"""High level plotting API within puma, to avoid code duplication."""

from __future__ import annotations

from puma.hlplots.aux_results import AuxResults
from puma.hlplots.n_track_origin import n_tracks_per_origin
from puma.hlplots.results import Results
from puma.hlplots.tagger import Tagger
from puma.hlplots.yutils import combine_suffixes, get_included_taggers

__all__ = [
    "AuxResults",
    "Results",
    "Tagger",
    "YumaConfig",
    "combine_suffixes",
    "get_included_taggers",
    "n_tracks_per_origin",
]



"""Unit test script for the functions in hlplots/tagger.py."""

from __future__ import annotations

import tempfile
import unittest
from pathlib import Path

import h5py
import numpy as np
from ftag import Flavours, get_mock_file
from ftag.hdf5 import structured_from_dict

from puma.hlplots import Results
from puma.hlplots.tagger import Tagger
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class ResultsTestCase(unittest.TestCase):
    """Test class for the Results class."""

    def test_set_signal_hcc(self):
        """Test set_signal for hcc."""
        results = Results(signal="hcc", sample="test")
        self.assertEqual(
            results.backgrounds,
            [Flavours.hbb, Flavours.top, Flavours.qcd],
        )
        self.assertEqual(results.signal, Flavours.hcc)

    def test_unsupported_signal(self):
        """Test usage of unsupported signal."""
        with self.assertRaises(ValueError):
            Results(signal="ujets", sample="test")

    def test_add_duplicated(self):
        """Test empty string as model name."""
        dummy_tagger_1 = Tagger("dummy")
        dummy_tagger_2 = Tagger("dummy")
        results = Results(signal="bjets", sample="test")
        results.add(dummy_tagger_1)
        with self.assertRaises(KeyError):
            results.add(dummy_tagger_2)

    def test_add_2_taggers(self):
        """Test empty string as model name."""
        dummy_tagger_1 = Tagger("dummy")
        dummy_tagger_2 = Tagger("dummy_2")
        results = Results(signal="bjets", sample="test")
        results.add(dummy_tagger_1)
        results.add(dummy_tagger_2)
        self.assertEqual(
            list(results.taggers.keys()),
            ["dummy (dummy)", "dummy_2 (dummy_2)"],  # pylint: disable=W0212
        )

    def test_get_taggers(self):
        """Test empty string as model name."""
        dummy_tagger_1 = Tagger("dummy")
        dummy_tagger_2 = Tagger("dummy_2")
        results = Results(signal="bjets", sample="test")
        results.add(dummy_tagger_1)
        results.add(dummy_tagger_2)
        retrieved_dummy_tagger_2 = results["dummy_2 (dummy_2)"]
        self.assertEqual(retrieved_dummy_tagger_2.name, dummy_tagger_2.name)

    def test_load_taggers_from_file(self):
        """Test for Results.load_taggers_from_file function."""
        fname = get_mock_file()[0]
        results = Results(signal="bjets", sample="test")
        taggers = [Tagger("MockTagger")]
        results.load_taggers_from_file(taggers, fname)
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_load_taggers_from_file_with_perf_vars(self):
        """Test for Results.load_taggers_from_file function."""
        fname = get_mock_file()[0]
        results = Results(signal="bjets", sample="test", perf_vars=["pt", "eta"])
        taggers = [Tagger("MockTagger")]
        results.load_taggers_from_file(taggers, fname)
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_add_taggers_with_cuts_override_perf_vars(self):
        """Test for Results.load_taggers_from_file function."""
        rng = np.random.default_rng(seed=16)
        cuts = [("eta", ">", 0)]
        tagger_cuts = [("pt", ">", 20)]
        fname = get_mock_file(num_jets=1000)[0]
        results = Results(signal="bjets", sample="test", perf_vars=["pt", "eta"])
        taggers = [Tagger("MockTagger", cuts=tagger_cuts)]

        results.load_taggers_from_file(
            taggers,
            fname,
            cuts=cuts,
            perf_vars={
                "pt": rng.exponential(100, size=1000),
                "eta": rng.normal(0, 1, size=1000),
            },
        )
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_add_taggers_with_cuts(self):
        fname = get_mock_file()[0]
        cuts = [("eta", ">", 0)]
        tagger_cuts = [("pt", ">", 20)]
        results = Results(signal="bjets", sample="test")
        taggers = [Tagger("MockTagger", cuts=tagger_cuts)]
        results.load_taggers_from_file(taggers, fname, cuts=cuts)
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_add_taggers_taujets(self):
        # get mock file and rename variables match taujets
        fname = get_mock_file()[0]
        results = Results(
            signal="bjets",
            sample="test",
            all_flavours=["bjets", "cjets", "ujets", "taujets"],
        )
        taggers = [Tagger("MockTagger", fxs={"fc": 0.1, "fb": 0.1, "ftau": 0.1})]
        results.load_taggers_from_file(taggers, fname)
        assert "MockTagger_ptau" in taggers[0].scores.dtype.names
        taggers[0].discriminant("bjets")

    def test_add_taggers_hbb(self):
        # get mock file and rename variables match hbb
        f = get_mock_file()[1]
        d = {}
        d["R10TruthLabel"] = f["jets"]["HadronConeExclTruthLabelID"]
        d["MockTagger_phbb"] = f["jets"]["MockTagger_pb"]
        d["MockTagger_phcc"] = f["jets"]["MockTagger_pc"]
        d["MockTagger_ptop"] = f["jets"]["MockTagger_pu"]
        d["MockTagger_pqcd"] = f["jets"]["MockTagger_pu"]
        d["pt"] = f["jets"]["pt"]
        array = structured_from_dict(d)
        with tempfile.TemporaryDirectory() as tmp_file:
            fname = Path(tmp_file) / "test.h5"
            with h5py.File(fname, "w") as f:
                f.create_dataset("jets", data=array)

            results = Results(signal="hbb", sample="test")
            results.load_taggers_from_file([Tagger("MockTagger")], fname, label_var="R10TruthLabel")

    def test_add_taggers_keep_nan(self):
        # get mock file and add nans
        f = get_mock_file()[1]
        d = {}
        d["HadronConeExclTruthLabelID"] = f["jets"]["HadronConeExclTruthLabelID"]
        d["MockTagger_pb"] = f["jets"]["MockTagger_pb"]
        d["MockTagger_pc"] = f["jets"]["MockTagger_pc"]
        d["MockTagger_pu"] = f["jets"]["MockTagger_pu"]
        d["pt"] = f["jets"]["pt"]
        n_nans = np.random.choice(range(100), 10)
        d["MockTagger_pb"][n_nans] = np.nan
        array = structured_from_dict(d)
        with tempfile.TemporaryDirectory() as tmp_file:
            fname = Path(tmp_file) / "test.h5"
            with h5py.File(fname, "w") as f:
                f.create_dataset("jets", data=array)

            results = Results(signal="bjets", sample="test", remove_nan=False)
            with self.assertRaises(ValueError):
                results.load_taggers_from_file([Tagger("MockTagger")], fname)

    def test_add_taggers_remove_nan(self):
        # get mock file and add nans
        f = get_mock_file()[1]
        d = {}
        d["HadronConeExclTruthLabelID"] = f["jets"]["HadronConeExclTruthLabelID"]
        d["MockTagger_pb"] = f["jets"]["MockTagger_pb"]
        d["MockTagger_pc"] = f["jets"]["MockTagger_pc"]
        d["MockTagger_pu"] = f["jets"]["MockTagger_pu"]
        d["pt"] = f["jets"]["pt"]
        n_nans = np.random.choice(range(100), 10, replace=False)
        d["MockTagger_pb"][n_nans] = np.nan
        array = structured_from_dict(d)
        with tempfile.TemporaryDirectory() as tmp_file:
            fname = Path(tmp_file) / "test.h5"
            with h5py.File(fname, "w") as f:
                f.create_dataset("jets", data=array)

            results = Results(signal="bjets", sample="test", remove_nan=True)
            with self.assertLogs("puma", "WARNING") as cm:
                results.load_taggers_from_file([Tagger("MockTagger")], fname)
            self.assertEqual(
                cm.output,
                [f"WARNING:puma:{len(n_nans)} NaN values found in loaded data." " Removing them."],
            )


class ResultsPlotsTestCase(unittest.TestCase):
    """Test class for the Results class running plots."""

    def setUp(self) -> None:
        """Set up for unit tests."""
        f = get_mock_file()[1]
        dummy_tagger_1 = Tagger("MockTagger")
        dummy_tagger_1.labels = np.array(
            f["jets"]["HadronConeExclTruthLabelID"],
            dtype=[("HadronConeExclTruthLabelID", "i4")],
        )
        dummy_tagger_1.scores = f["jets"]
        dummy_tagger_1.label = "dummy tagger"
        self.dummy_tagger_1 = dummy_tagger_1

    def test_plot_probs_bjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_probs()
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_discs_bjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_discs()
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_discs_cjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fb": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="cjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_discs(wp_vlines=[60])
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_roc_bjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_rocs()
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_roc_cjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fb": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="cjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_rocs(fontsize=5)
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_var_perf_err(self):
        """Tests the performance plots throws errors with invalid inputs."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.disc_cut = 2
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            with self.assertRaises(ValueError):
                results.plot_var_perf(
                    bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                )
            with self.assertRaises(ValueError):
                results.plot_var_perf(
                    bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                    disc_cut=1,
                    working_point=0.5,
                )

    def test_plot_var_eff_per_flat_rej_err(self):
        """Tests the performance vs flat rejection plots throws errors
        with invalid inputs.
        """
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.disc_cut = 2
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file, self.assertRaises(ValueError):
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.plot_flat_rej_var_perf(
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                fixed_rejections={"cjets": 10, "ujets": 100},
                working_point=0.5,
            )
        with tempfile.TemporaryDirectory() as tmp_file, self.assertRaises(ValueError):
            results.plot_flat_rej_var_perf(
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                fixed_rejections={"cjets": 10, "ujets": 100},
                disc_cut=0.5,
            )

    def test_plot_var_perf_bjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.disc_cut = 2
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_var_perf(
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                working_point=0.7,
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_var_perf_extra_kwargs(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.disc_cut = 2
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_var_perf(
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                working_point=0.7,
                y_scale=1.3,
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_var_perf_multi_bjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.disc_cut = 2
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores)),
            "eta": rng.normal(0, 1, size=len(self.dummy_tagger_1.scores)),
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_var_perf(
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                working_point=0.7,
                perf_var="pt",
            )
            results.plot_var_perf(
                bins=np.linspace(-0.5, 0.5, 10), working_point=0.7, perf_var="eta"
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_var_perf_cjets(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fb": 0.05}
        self.dummy_tagger_1.working_point = 0.5
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="cjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_var_perf(
                h_line=self.dummy_tagger_1.working_point,
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                working_point=0.7,
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_beff_vs_flat_rej(self):
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.working_point = 0.5
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_flat_rej_var_perf(
                fixed_rejections={"cjets": 10, "ujets": 100},
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                h_line=0.5,
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_beff_vs_flat_rej_extra_kwargs(self):
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        self.dummy_tagger_1.working_point = 0.5
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_flat_rej_var_perf(
                fixed_rejections={"cjets": 10, "ujets": 100},
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
                h_line=0.5,
                y_scale=1.3,
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_ceff_vs_flat_rej(self):
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fb": 0.05}
        self.dummy_tagger_1.working_point = 0.5
        rng = np.random.default_rng(seed=16)
        self.dummy_tagger_1.perf_vars = {
            "pt": rng.exponential(100, size=len(self.dummy_tagger_1.scores))
        }
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="cjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_flat_rej_var_perf(
                fixed_rejections={"bjets": 10, "ujets": 100},
                bins=[20, 30, 40, 60, 85, 110, 140, 175, 250],
            )
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_fraction_scans_hbb_error(self):
        """Test that correct error is raised."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="hbb", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            with self.assertRaises(ValueError):
                results.plot_fraction_scans(rej=False)

    def test_plot_fraction_scans_bjets_eff(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_fraction_scans(rej=False, optimal_fc=True)
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_fraction_scans_cjets_rej(self):
        """Test that png file is being created."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fb": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="cjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            results.plot_fraction_scans(rej=True, optimal_fc=True)
            for fpath in results.saved_plots:
                assert fpath.is_file()
            results.saved_plots = []

    def test_plot_fraction_scans_multiple_bkg_error(self):
        """Test error of more than two backgrounds."""
        self.dummy_tagger_1.reference = True
        self.dummy_tagger_1.fxs = {"fc": 0.05}
        with tempfile.TemporaryDirectory() as tmp_file:
            results = Results(signal="bjets", sample="test", output_dir=tmp_file)
            results.add(self.dummy_tagger_1)
            with self.assertRaises(ValueError):
                results.plot_fraction_scans(
                    rej=False,
                    optimal_fc=True,
                    backgrounds=[Flavours.cjets, Flavours.ujets, Flavours.taujets],
                )

    def test_make_plot_error(self):
        """Test error of non-existing plot type."""
        results = Results(signal="bjets", sample="test")
        with self.assertRaises(ValueError):
            results.make_plot(plot_type="crash", kwargs={})



from __future__ import annotations

import math as m

import numpy as np
from matplotlib import pyplot as plt

from puma.plot_base import PlotBase
from puma.utils import logger


class MatshowPlot(PlotBase):
    """Plot Matrix class."""

    def __init__(
        self,
        x_ticklabels: list | None = None,
        x_ticks_rotation: int = 90,
        y_ticklabels: list | None = None,
        show_entries: bool = True,
        show_percentage: bool = False,
        text_color_threshold: float = 0.408,
        colormap: plt.cm = plt.cm.Oranges,
        show_cbar: bool = True,
        cbar_label: str | None = None,
        **kwargs,
    ) -> None:
        """Plot a matrix with matplotlib matshow.

        Parameters
        ----------
        x_ticklabels : list | None, optional
            Names of the matrix's columns; if None, indices are shown. by default None
        x_ticks_rotation : int, optional
            Rotation of the columns' names, by default 90
        y_ticklabels : list | None, optional
            Names of the matrix's rows; if None, indices are shown. by default None
        show_entries : bool, optional
            If True, show matrix entries as numbers in the matrix pixels. by default True
        show_percentage : bool, optional
            If True, if matrix entries are percentages (i.e. numbers in [0,1]), format them as
            percentages. by default False
        text_color_threshold : float, optional
            threshold on the relative luminance of the colormap bkg color after which the text color
            switches to black, to allow better readability on lighter cmap bkg colors.
            If 1, all text is white; if 0, all text is black. by default 0.408
        colormap : plt.cm, optional
            Colormap for the plot, by default `plt.cm.Oranges`
        show_cbar : bool, optional
            Whether to plot the colorbar or not, by default True
        cbar_label : str | None, optional
            Label of the colorbar, by default None
        **kwargs : kwargs
            Keyword arguments for `puma.PlotObject`

        Example
        -------
        >>> matrix_plotter = MatshowPlot()
        >>> mat = np.random.rand(4, 3)
        >>> matrix_plotter.draw(mat)
        """
        super().__init__(**kwargs)

        self.x_ticklabels = x_ticklabels
        self.x_ticks_rotation = x_ticks_rotation
        self.y_ticklabels = y_ticklabels
        self.show_entries = show_entries
        self.show_percentage = show_percentage
        self.text_color_threshold = text_color_threshold
        self.colormap = colormap
        self.show_cbar = show_cbar
        self.cbar_label = cbar_label

        # Specifying figsize if not specified by user
        if self.figsize is None:
            self.figsize = (10, 10.5)
        self.initialise_figure()

    def __get_luminance(self, rgbaColor):
        """Calculate the relative luminance of a color according to W3C standards.
        For the details of the conversion see: https://www.w3.org/WAI/GL/wiki/Relative_luminance .

        Parameters
        ----------
        rgbColor : tuple
            (r,g,b,a) color (returned from `plt.cm` colormap)

        Returns
        -------
        float
            Relative luminance of the color.
        """
        # Converting to np.ndarray, ignoring alpha channel
        rgbaColor = np.array(rgbaColor[:-1])
        rgbaColor = np.where(
            rgbaColor <= 0.03928, rgbaColor / 12.92, ((rgbaColor + 0.055) / 1.055) ** 2.4
        )
        weights = np.array([0.2126, 0.7152, 0.0722])
        return np.dot(rgbaColor, weights)

    def __plot(self, matrix):
        """Plot the Matrix."""
        n_cols = matrix.shape[1]
        n_rows = matrix.shape[0]

        im = self.axis_top.matshow(matrix, cmap=self.colormap)

        # If mat entries have to be plotted
        if self.show_entries:
            # Mapping mat values in [0,1], as it's done by matplotlib
            # to associate them to the colors of the colormap
            normMat = matrix - np.min(matrix)
            normMat /= np.max(matrix) - np.min(matrix)

            # Adding text values in the matrix pixels
            for i in range(n_rows):
                for j in range(n_cols):
                    # Choosing the text color: black if color is light, white if color is dark
                    # Getting the bkg color from the cmap
                    color = self.colormap(normMat[i, j])
                    # Calculating the bkg relative luminance
                    luminance = self.__get_luminance(color)
                    # Choosing the appropriate color
                    color = "white" if luminance <= self.text_color_threshold else "black"

                    # If matrix entry is an int, do not show decimals
                    if m.modf(matrix[i, j])[0] == 0:
                        text = f"{matrix[i, j]:.0f}"
                    # Else, round it or show it as percentage
                    else:
                        text = (
                            f"{matrix[i, j]:.3f}"
                            if not self.show_percentage
                            else f"{matrix[i, j] * 100:.0f}%"
                        )
                    # Plotting text
                    self.axis_top.text(
                        x=j,
                        y=i,
                        s=text,
                        va="center",
                        ha="center",
                        c=color,
                        fontsize=self.fontsize,
                    )
        if self.show_cbar:
            # Plotting colorbar
            cbar = self.fig.colorbar(im)
            # If using percentages, converting cbar labels to percentages
            if self.show_entries and self.show_percentage:
                minMat = np.min(matrix)
                maxMat = np.max(matrix)
                cbar.set_ticks(
                    ticks=np.linspace(minMat, maxMat, 5),
                    labels=[f"{i}%" for i in np.round(np.linspace(minMat, maxMat, 5) * 100, 2)],
                    fontsize=self.fontsize,
                )
            if self.cbar_label is not None:
                cbar.ax.set_ylabel(self.cbar_label, fontsize=self.fontsize)

        # Setting tick labels
        if self.x_ticklabels is None:
            self.x_ticklabels = [str(i) for i in range(n_cols)]
            logger.info("MatshowPlot: no x_ticklabels given, using indices instead.")
        if self.y_ticklabels is None:
            self.y_ticklabels = [str(i) for i in range(n_rows)]
            logger.info("MatshowPlot: no y_ticklabels given, using indices instead.")

        # Writing class names on the axes
        positions = list(range(n_cols))
        self.axis_top.set_xticks(
            np.array(positions) + 0.25,
            labels=self.x_ticklabels,
            rotation=self.x_ticks_rotation,
            fontsize=self.fontsize,
            ha="right",
        )
        positions = list(range(n_rows))
        self.axis_top.set_yticks(positions, labels=self.y_ticklabels, fontsize=self.fontsize)
        # Put xticks to the bottom
        self.axis_top.xaxis.tick_bottom()

        # Finished plotting, can apply atlas_style
        self.plotting_done = True
        # Applying atlas style
        if self.apply_atlas_style:
            # Apply ATLAS style
            self.atlasify()
            # Mirror y axis to have the diagonal in the common orientation
            self.axis_top.invert_yaxis()

        # Disable x and y ticks for better appearance
        self.axis_top.tick_params("x", which="both", top=False, bottom=False)
        self.axis_top.tick_params("y", which="both", right=False, left=False)
        # Disable grid for better appearance
        self.axis_top.grid(False)

        # Setting title and label
        self.set_xlabel()
        self.set_ylabel(self.axis_top)
        self.set_title()

    def draw(self, matrix):
        """Draw a matrix with the class customized appearance.

        Parameters
        ----------
        matrix : np.ndarray
            The matrix to be plotted.
        """
        # Checking size consistency of ticklabels
        if self.x_ticklabels is not None:
            assert (
                len(self.x_ticklabels) == matrix.shape[1]
            ), "MatshowPlot: mismatch between x_tickslabels and number of columns in the matrix."

        if self.y_ticklabels is not None:
            assert (
                len(self.y_ticklabels) == matrix.shape[0]
            ), "MatshowPlot: mismatch between y_tickslabels and number of rows in the matrix."

        self.__plot(matrix)



"""Unit test script for the functions in metrics.py."""

from __future__ import annotations

import unittest

import numpy as np

from puma.metrics import calc_eff, calc_rej, calc_separation, eff_err, rej_err
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class SeparationTestCase(unittest.TestCase):
    """Test class for the puma.metrics.calc_separation."""

    def setUp(self):
        """Define a default (seeded) random number generator for all tests."""
        self.rng = np.random.default_rng(42)

    def test_equal_datasets(self):
        """Separation of two equal datasets (=0)."""
        values_a = np.array([1, 1, 2, 2])
        self.assertEqual(0, calc_separation(values_a, values_a)[0])

    def test_completely_separated(self):
        """Separation of two completely separated distributions."""
        values_a = np.array([0.1, 0.5, 0.8, 1])
        values_b = np.array([1.1, 1.5, 1.8, 2])
        self.assertAlmostEqual(1, calc_separation(values_a, values_b)[0])

    def test_completely_separated_bad_binning(self):
        """Separation of two completely separated distributions if the binning
        if chosen such that they share one bin
        .
        """
        values_a = np.array([0.1, 0.5, 0.8, 1])
        values_b = np.array([1.1, 1.5, 1.8, 2])
        self.assertNotAlmostEqual(1, calc_separation(values_a, values_b, bins=3)[0])

    def test_half_separated(self):
        """Separation of 0.5."""
        values_a = np.array([0, 1])
        values_b = np.array([1, 2])
        self.assertAlmostEqual(0.5, calc_separation(values_a, values_b, bins=3)[0])

    def test_return_bins(self):
        """Test if bins are correctly returned."""
        values_a = np.array([0, 1])
        values_b = np.array([1, 2])

        _, _, hist_a, hist_b, bin_edges = calc_separation(
            values_a,
            values_b,
            bins=3,
            return_hist=True,
        )
        # Check for correct values in hist_a, hist_b and bin_edges
        np.testing.assert_array_equal(np.array([0.5, 0.5, 0]), hist_a)
        np.testing.assert_array_equal(np.array([0, 0.5, 0.5]), hist_b)
        np.testing.assert_array_equal(np.array([0, 2 / 3, 4 / 3, 2]), bin_edges)

    def test_bins_range(self):
        """Test if bins_range is properly treated."""
        values_a = np.array([0, 1])
        values_b = np.array([1, 2])

        _, _, hist_a, _, bin_edges = calc_separation(
            values_a,
            values_b,
            bins=4,
            bins_range=(0, 4),  # this should result in bin edges [0, 1, 2, 3, 4]
            return_hist=True,
        )
        # Check for correct values in hist_a and bin_edges
        np.testing.assert_array_equal(np.array([0.5, 0.5, 0, 0]), hist_a)
        np.testing.assert_array_equal(np.array([0, 1, 2, 3, 4]), bin_edges)


class CalcEffAndRejTestCase(unittest.TestCase):
    """Test class for the puma.metrics functions."""

    def setUp(self):
        rng = np.random.default_rng(seed=42)
        self.disc_sig = rng.normal(loc=3, size=100_000)
        self.disc_bkg = rng.normal(loc=0, size=100_000)

    def test_float_target(self):
        """Test efficiency and cut value calculation for one target value."""
        # we target a signal efficiency of 0.841345, which is the integral of a gaussian
        # from μ-1o to infinity
        # -->   the cut_value should be at 2, since the signal is a normal distribution
        #       with mean 3
        # https://www.wolframalpha.com/input?i=integrate+1%2Fsqrt%282+pi%29+*+exp%28-0.5*%28x-3%29**2%29+from+2+to+oo
        # -->   For the bkg efficiency this means that we integrate a normal distr.
        #       from μ+2o to infinity --> expect a value of 0.0227501
        # https://www.wolframalpha.com/input?i=integrate+1%2Fsqrt%282+pi%29+*+exp%28-0.5*x**2%29+from+2+to+oo
        bkg_eff, cut = calc_eff(self.disc_sig, self.disc_bkg, target_eff=0.841345, return_cuts=True)
        # the values here differ slightly from the values of the analytical integral,
        # since we use random numbers
        self.assertAlmostEqual(cut, 1.9956997)
        self.assertAlmostEqual(bkg_eff, 0.02367)

        # same for rejection, just use rej = 1 / eff
        bkg_rej, cut = calc_rej(self.disc_sig, self.disc_bkg, target_eff=0.841345, return_cuts=True)
        self.assertAlmostEqual(cut, 1.9956997)
        self.assertAlmostEqual(bkg_rej, 1 / 0.02367)

    def test_array_target(self):
        """Test efficiency and cut value calculation for list of target efficiencies."""
        # explanation is the same as above, now also cut the signal in the middle
        # --> target sig.efficiency 0.841345 and 0.5 --> cut at 2 and 3
        bkg_eff, cut = calc_eff(
            self.disc_sig, self.disc_bkg, target_eff=[0.841345, 0.5], return_cuts=True
        )
        # the values here differ slightly from the values of the analytical integral,
        # since we use random numbers
        np.testing.assert_array_almost_equal(cut, np.array([1.9956997, 2.990996]))
        np.testing.assert_array_almost_equal(bkg_eff, np.array([0.02367, 0.00144]))

        # same for rejection, just use rej = 1 / eff
        bkg_rej, cut = calc_rej(
            self.disc_sig, self.disc_bkg, target_eff=[0.841345, 0.5], return_cuts=True
        )
        np.testing.assert_array_almost_equal(cut, np.array([1.9956997, 2.990996]))
        np.testing.assert_array_almost_equal(bkg_rej, 1 / np.array([0.02367, 0.00144]))


class EffErrTestCase(unittest.TestCase):
    """Test class for the puma.metrics functions."""

    def test_zero_n_case(self):
        """Test eff_err function."""
        with self.assertRaises(ValueError):
            eff_err(0, 0)

    def test_negative_n_case(self):
        """Test eff_err function."""
        with self.assertRaises(ValueError):
            eff_err(0, -1)

    def test_one_case(self):
        """Test eff_err function."""
        self.assertEqual(eff_err(1, 1), 0)

    def test_example_case(self):
        """Test eff_err function."""
        x_eff = np.array([0.25, 0.5, 0.75])
        error_eff = np.array([0.043301, 0.05, 0.043301])
        np.testing.assert_array_almost_equal(eff_err(x_eff, 100), error_eff)


class RejErrTestCase(unittest.TestCase):
    """Test class for the puma.metrics functions."""

    def setUp(self):
        self.array = np.array([[1, 2, 3], [4, 5, 6]])

    def test_zero_n_case(self):
        """Test rej_err function."""
        with self.assertRaises(ValueError):
            rej_err(0, 0)

    def test_negative_n_case(self):
        """Test rej_err function."""
        with self.assertRaises(ValueError):
            rej_err(0, -1)

    def test_one_case(self):
        """Test rej_err function."""
        self.assertEqual(rej_err(1, 1), 0)

    def test_zero_x_case(self):
        """Test rej_err function."""
        with self.assertRaises(ValueError):
            rej_err(np.array([0, 1, 2]), 3)

    def test_example_case(self):
        """Test rej_err function."""
        x_rej = np.array([20, 50, 100])
        error_rej = np.array([8.717798, 35.0, 99.498744])
        np.testing.assert_array_almost_equal(rej_err(x_rej, 100), error_rej)



"""Tools for metrics module."""

from __future__ import annotations

import numpy as np
from scipy.ndimage import gaussian_filter1d

from puma.utils import logger
from puma.utils.histogram import hist_w_unc, save_divide


def weighted_percentile(
    data: np.ndarray,
    perc: np.ndarray,
    weights: np.ndarray = None,
):
    """Calculate weighted percentile.

    Implementation according to https://stackoverflow.com/a/29677616/11509698
    (https://en.wikipedia.org/wiki/Percentile#The_weighted_percentile_method)

    Parameters
    ----------
    data : np.ndarray
        Data array
    perc : np.ndarray
        Percentile array
    weights : np.ndarray
        Weights array, by default None

    Returns
    -------
    np.ndarray
        Weighted percentile array
    """
    if weights is None:
        weights = np.ones_like(data)
    dtype = np.float64 if np.sum(weights) > 1000000 else np.float32
    ix = np.argsort(data)
    data = data[ix]  # sort data
    weights = weights[ix]  # sort weights
    cdf = np.cumsum(weights, dtype=dtype) - 0.5 * weights
    cdf -= cdf[0]
    cdf /= cdf[-1]
    return np.interp(perc, cdf, data)


def calc_eff(
    sig_disc: np.ndarray,
    bkg_disc: np.ndarray,
    target_eff: float | list | np.ndarray,
    return_cuts: bool = False,
    sig_weights: np.ndarray = None,
    bkg_weights: np.ndarray = None,
):
    """Calculate efficiency.

    Parameters
    ----------
    sig_disc : np.ndarray
        Signal discriminant
    bkg_disc : np.ndarray
        Background discriminant
    target_eff : float or list or np.ndarray
        Working point which is used for discriminant calculation
    return_cuts : bool
        Specifies if the cut values corresponding to the provided WPs are returned.
        If target_eff is a float, only one cut value will be returned. If target_eff
        is an array, target_eff is an array as well.
    sig_weights : np.ndarray
        Weights for signal events
    bkg_weights : np.ndarray
        Weights for background events

    Returns
    -------
    eff : float or np.ndarray
        Efficiency.
        Return float if target_eff is a float, else np.ndarray
    cutvalue : float or np.ndarray
        Cutvalue if return_cuts is True.
        Return float if target_eff is a float, else np.ndarray
    """
    # float | np.ndarray for both target_eff and the returned values
    return_float = False
    if isinstance(target_eff, float):
        return_float = True

    target_eff = np.asarray([target_eff]).flatten()

    cutvalue = weighted_percentile(sig_disc, 1.0 - target_eff, weights=sig_weights)
    sorted_args = np.argsort(1 - target_eff)  # need to sort the cutvalues to get the correct order
    hist, _ = np.histogram(bkg_disc, (-np.inf, *cutvalue[sorted_args], np.inf), weights=bkg_weights)
    eff = hist[::-1].cumsum()[-2::-1] / hist.sum()
    eff = eff[sorted_args]

    if return_float:
        eff = eff[0]
        cutvalue = cutvalue[0]

    if return_cuts:
        return eff, cutvalue
    return eff


def calc_rej(
    sig_disc: np.ndarray,
    bkg_disc: np.ndarray,
    target_eff,
    return_cuts: bool = False,
    sig_weights: np.ndarray = None,
    bkg_weights: np.ndarray = None,
    smooth: bool = False,
):
    """Calculate efficiency.

    Parameters
    ----------
    sig_disc : np.ndarray
        Signal discriminant
    bkg_disc : np.ndarray
        Background discriminant
    target_eff : float or list
        Working point which is used for discriminant calculation
    return_cuts : bool
        Specifies if the cut values corresponding to the provided WPs are returned.
        If target_eff is a float, only one cut value will be returned. If target_eff
        is an array, target_eff is an array as well.
    sig_weights : np.ndarray
        Weights for signal events, by default None
    bkg_weights : np.ndarray
        Weights for background events, by default None

    Returns
    -------
    rej : float or np.ndarray
        Rejection.
        If target_eff is a float, a float is returned if it's a list a np.ndarray
    cut_value : float or np.ndarray
        Cutvalue if return_cuts is True.
        If target_eff is a float, a float is returned if it's a list a np.ndarray
    """
    # float | np.ndarray for both target_eff and the returned values
    eff = calc_eff(
        sig_disc=sig_disc,
        bkg_disc=bkg_disc,
        target_eff=target_eff,
        return_cuts=return_cuts,
        sig_weights=sig_weights,
        bkg_weights=bkg_weights,
    )
    rej = save_divide(1, eff[0] if return_cuts else eff, np.inf)
    if smooth:
        rej = gaussian_filter1d(rej, sigma=1, radius=2, mode="nearest")
    if return_cuts:
        return rej, eff[1]
    return rej


def eff_err(
    arr: np.ndarray,
    n_counts: int,
    suppress_zero_divison_error: bool = False,
    norm: bool = False,
) -> np.ndarray:
    """Calculate statistical efficiency uncertainty.

    Parameters
    ----------
    arr : numpy.array
        Efficiency values
    n_counts : int
        Number of used statistics to calculate efficiency
    suppress_zero_divison_error : bool
        Not raising Error for zero division
    norm : bool, optional
        If True, normed (relative) error is being calculated, by default False

    Returns
    -------
    numpy.array
        Efficiency uncertainties

    Raises
    ------
    ValueError
        If n_counts <=0

    Notes
    -----
    This method uses binomial errors as described in section 2.2 of
    https://inspirehep.net/files/57287ac8e45a976ab423f3dd456af694
    """
    logger.debug("Calculating efficiency error.")
    logger.debug("arr: %s", arr)
    logger.debug("n_counts: %i", n_counts)
    logger.debug("suppress_zero_divison_error: %s", suppress_zero_divison_error)
    logger.debug("norm: %s", norm)
    if np.any(n_counts <= 0) and not suppress_zero_divison_error:
        raise ValueError(f"You passed as argument `N` {n_counts} but it has to be larger 0.")
    if norm:
        return np.sqrt(arr * (1 - arr) / n_counts) / arr
    return np.sqrt(arr * (1 - arr) / n_counts)


def rej_err(
    arr: np.ndarray,
    n_counts: int,
    norm: bool = False,
) -> np.ndarray:
    """Calculate the rejection uncertainties.

    Parameters
    ----------
    arr : numpy.array
        Rejection values
    n_counts : int
        Number of used statistics to calculate rejection
    norm : bool, optional
        If True, normed (relative) error is being calculated, by default False

    Returns
    -------
    numpy.array
        Rejection uncertainties

    Raises
    ------
    ValueError
        If n_counts <=0
    ValueError
        If any rejection value is 0

    Notes
    -----
    Special case of `eff_err()`
    """
    logger.debug("Calculating rejection error.")
    logger.debug("arr: %s", arr)
    logger.debug("n_counts: %i", n_counts)
    logger.debug("norm: %s", norm)
    if np.any(n_counts <= 0):
        raise ValueError(f"You passed as argument `n_counts` {n_counts} but it has to be larger 0.")
    if np.any(arr == 0):
        raise ValueError("One rejection value is 0, cannot calculate error.")
    if norm:
        return np.power(arr, 2) * eff_err(1 / arr, n_counts) / arr
    return np.power(arr, 2) * eff_err(1 / arr, n_counts)


def calc_separation(
    values_a: np.ndarray,
    values_b: np.ndarray,
    bins: int = 100,
    bins_range: tuple | None = None,
    return_hist: bool = False,
) -> float:
    """Calculates the separation of two distributions.

    Parameters
    ----------
    values_a : np.ndarray
        First distribution
    values_b : np.ndarray
        Second distribution
    bins : int, optional
        Number of bins used for the histograms (common binning over whole range if
        `bins_range` is not defined), by default 100
    bins_range : tuple, optional
        Lower and upper limit for binning. If not provided, the whole range of the
        joined distribution is used, by default None
    return_hist : bool, optional
        Option to also return the hist_a, hist_b and bin_edges arrays

    Returns
    -------
    float
        separation value
    float
        separation uncertainty
    numpy.ndarray
        Bins height of histogram a (only returned if `return_hist` is True)
    numpy.ndarray
        Bins height of histogram b (only returned if `return_hist` is True)
    numpy.ndarray
        Bin edges of the two histograms (only returned if `return_hist` is True)
    """
    _, bin_edges = np.histogram(np.hstack([values_a, values_b]), bins=bins, range=bins_range)

    _, hist_a, unc_a, _ = hist_w_unc(values_a, bin_edges)
    _, hist_b, unc_b, _ = hist_w_unc(values_b, bin_edges)

    separation = 0.5 * np.sum(
        save_divide(
            numerator=(hist_a - hist_b) ** 2,
            denominator=(hist_a + hist_b),
            default=0,
        )
    )

    # helper variable for error calculation below
    ab_diff_over_sum = save_divide(hist_a - hist_b, hist_a + hist_b, default=0)

    separation_uncertainty = 0.5 * np.sqrt(
        np.sum(
            (unc_a * (2 * ab_diff_over_sum - ab_diff_over_sum**2)) ** 2
            + (unc_b * (-2 * ab_diff_over_sum - ab_diff_over_sum**2)) ** 2
        )
    )

    if return_hist:
        return separation, separation_uncertainty, hist_a, hist_b, bin_edges
    return separation, separation_uncertainty
